"""
Search Implementation: Reserve California
"""

import logging
import sys
from abc import ABC, abstractmethod
from datetime import timedelta
from typing import Any, Dict, List, Optional, Type, Union

from dateutil.relativedelta import relativedelta

from camply.containers import AvailableCampsite, RecreationArea, SearchWindow
from camply.containers.data_containers import ListedCampsite
from camply.providers.usedirect.variations import (
    AlabamaStateParks,
    ArizonaStateParks,
    FairfaxCountyParks,
    FloridaStateParks,
    MaricopaCountyParks,
    MinnesotaStateParks,
    MissouriStateParks,
    NorthernTerritory,
    OhioStateParks,
    OregonMetro,
    ReserveCalifornia,
    VirginiaStateParks,
)
from camply.search.base_search import BaseCampingSearch
from camply.utils import logging_utils, make_list
from camply.utils.logging_utils import format_log_string, log_sorted_response

logger = logging.getLogger(__name__)


class SearchUseDirect(BaseCampingSearch, ABC):
    """
    Searches on UseDirect.com for Campsites
    """

    @property
    @abstractmethod
    def provider_class(self) -> Type[BaseCampingSearch]:
        """
        Provider Class to be used for Search
        """
        pass

    def __init__(
        self,
        search_window: Union[SearchWindow, List[SearchWindow]],
        recreation_area: List[int],
        weekends_only: bool = False,
        campgrounds: Optional[Union[List[str], str]] = None,
        nights: int = 1,
        **kwargs,
    ) -> None:
        """
        Initialize with Search Parameters

        Parameters
        ----------
        search_window: Union[SearchWindow, List[SearchWindow]]
            Search Window tuple containing start date and End Date
        recreation_area: List[int]
            The IDs of the recreation area to be searched.
        weekends_only: bool
            Whether to only search for Camping availabilities on the weekends (Friday /
            Saturday nights)
        campgrounds: Union[List[int], int]
            Campground ID or List of Campground IDs
        nights: int
            minimum number of consecutive nights to search per campsite,defaults to 1
        """
        super().__init__(
            search_window=search_window,
            weekends_only=weekends_only,
            nights=nights,
            **kwargs,
        )
        self._recreation_area_ids: List[int] = make_list(recreation_area, coerce=int)
        self._campground_ids: List[int] = make_list(campgrounds, coerce=int)
        campsites = make_list(kwargs.get("campsites", []), coerce=int) or []
        if len(campsites) > 0:
            self.campsite_finder.validate_campsites(
                campsites=campsites, facility_ids=self._campground_ids
            )
        try:
            assert any(
                [
                    self._campground_ids not in (None, []),
                    self._recreation_area_ids not in (None, []),
                ]
            )
        except AssertionError:
            logger.error(
                f"You must provide a Campground ID or a Recreation Area ID to {self.provider_class.__name__}"
            )
            sys.exit(1)
        if self._campground_ids:
            self.campgrounds = self.campsite_finder.find_campgrounds(
                campground_id=self._campground_ids,
                verbose=False,
            )
        else:
            self.campgrounds = self.campsite_finder.find_campgrounds(
                rec_area_id=self._recreation_area_ids,
                verbose=False,
            )
        self.campground_ids = [item.facility_id for item in self.campgrounds]
        if len(self.campground_ids) == 0:
            logger.error("No Campsites Found Matching Your Search Criteria")
            sys.exit(1)
        if kwargs.get("equipment", ()):
            logger.warning(
                "%s Doesn't Support Equipment, yet 🙂", self.provider_class.__name__
            )

    def get_all_campsites(self, **kwargs: Dict[str, Any]) -> List[AvailableCampsite]:
        """
        Retrieve All Campsites from the UseDirect API

        Parameters
        ----------
        kwargs: Dict[str, Any]

        Returns
        -------
        List[AvailableCampsite]
        """
        logger.info(f"Searching across {len(self.campgrounds)} campgrounds")
        for campground in self.campgrounds:
            log_str = format_log_string(campground)
            logger.info("    %s", log_str)
        campsites_found: List[AvailableCampsite] = []
        for month in self.search_months:
            for campground in self.campgrounds:
                logger.info(
                    f"Searching {campground.facility_name}, {campground.recreation_area} "
                    f"({campground.facility_id}) for availability: "
                    f"{month.strftime('%B, %Y')}"
                )
                end_date = month + relativedelta(months=1) - timedelta(days=1)
                campsites = self.campsite_finder.get_campsites(
                    campground_id=campground.facility_id,
                    start_date=month,
                    end_date=end_date,
                )
                logger.info(
                    f"\t{logging_utils.get_emoji(campsites)}\t"
                    f"{len(campsites)} total sites found in month of "
                    f"{month.strftime('%B')}"
                )
                campsites_found += campsites
        campsite_df = self.campsites_to_df(campsites=campsites_found)
        campsite_df_validated = self._filter_date_overlap(campsites=campsite_df)
        consolidated_campsites = self._consolidate_campsites(
            campsite_df=campsite_df_validated, nights=self.nights
        )
        compiled_campsites = self.df_to_campsites(campsite_df=consolidated_campsites)
        return compiled_campsites

    @classmethod
    def find_recreation_areas(
        cls, search_string: str, **kwargs
    ) -> List[RecreationArea]:
        """
        Return the UseDirect Recreation Areas
        """
        rec_areas = cls.provider_class().search_for_recreation_areas(
            query=search_string, state=kwargs.get("state")
        )
        logger.info(f"{len(rec_areas)} Matching Recreation Areas Found")
        log_sorted_response(rec_areas)
        return rec_areas

    def list_campsite_units(self) -> List[ListedCampsite]:
        """
        List Campsite Units

        Returns
        -------
        List[ListedCampsite]
        """
        if not self.campsite_finder.usedirect_campsites:
            self.campsite_finder.get_campsite_metadata(facility_ids=self.campground_ids)
        sorted_campsites = sorted(
            self.campsite_finder.usedirect_campsites.values(),
            key=lambda x: x.OrderByRaw,
        )
        logged_campsites = [
            ListedCampsite(id=item.UnitId, name=item.Name, facility_id=item.FacilityId)
            for item in sorted_campsites
        ]
        self.log_listed_campsites(
            campsites=logged_campsites,
            facilities=self.campgrounds,
        )
        return logged_campsites


class SearchReserveCalifornia(SearchUseDirect):
    """
    Search ReserveCalifornia
    """

    provider_class = ReserveCalifornia


class SearchNorthernTerritory(SearchUseDirect):
    """
    Searches the Australian Northern Territory for Campsites
    """

    provider_class = NorthernTerritory


class SearchFloridaStateParks(SearchUseDirect):
    """
    Searches on FloridaStateParks.org for Campsites
    """

    provider_class = FloridaStateParks


class SearchOregonMetro(SearchUseDirect):
    """
    Searches on OregonMetro.gov for Campsites (Portland Metro)
    """

    provider_class = OregonMetro


class SearchOhioStateParks(SearchUseDirect):
    """
    Searches on ReserveOhio.com for Campsites
    """

    provider_class = OhioStateParks


class SearchVirginiaStateParks(SearchUseDirect):
    """
    Searches on ReserveVAParks.com for Campsites
    """

    provider_class = VirginiaStateParks


class SearchArizonaStateParks(SearchUseDirect):
    """
    Searches on AZStateParks.com for Campsites
    """

    provider_class = ArizonaStateParks


class SearchMaricopaCountyParks(SearchUseDirect):
    """
    Searches on MaricopaCountyParks.org for Campsites (Arizona)
    """

    provider_class = MaricopaCountyParks


class SearchMissouriStateParks(SearchUseDirect):
    """
    Searches on icampmo1.usedirect.com for Campsites
    """

    provider_class = MissouriStateParks


class SearchAlabamaStateParks(SearchUseDirect):
    """
    Searches on ReserveAlaPark.com for Campsites
    """

    provider_class = AlabamaStateParks


class SearchFairfaxCountyParks(SearchUseDirect):
    """
    Searches on fairfax.usedirect.com for Campsites (Virginia)
    """

    provider_class = FairfaxCountyParks


class SearchMinnesotaStateParks(SearchUseDirect):
    """
    Searches on ReserveMN.usedirect.com for Campsites
    """

    provider_class = MinnesotaStateParks



"""
Usedirect specific variations
"""

from camply.providers.usedirect import UseDirectProvider

# State Parks


class AlabamaStateParks(UseDirectProvider):
    """
    Alabama State Parks
    """

    base_url = "https://alabamardr.usedirect.com"
    campground_url = "https://www.reservealapark.com"
    rdr_path = "alabamardr"
    booking_path = "AlabamaWebHome/Facilities/SearchViewUnitAvailabity.aspx"
    booking_path_params = False
    state_code = "AL"


class ReserveCalifornia(UseDirectProvider):
    """
    ReserveCalifornia
    """

    base_url = "https://calirdr.usedirect.com"
    campground_url = "https://www.reservecalifornia.com"
    state_code = "CA"


class FloridaStateParks(UseDirectProvider):
    """
    Florida State Parks
    """

    base_url = "https://floridardr.usedirect.com"
    campground_url = "https://www.reserve.floridastateparks.org"
    rdr_path = "FloridaRDR"
    booking_path = "Web"
    state_code = "FL"


class OhioStateParks(UseDirectProvider):
    """
    Ohio State Parks
    """

    base_url = "https://ohiordr.usedirect.com"
    campground_url = "https://www.OhioStateParks.com"
    rdr_path = "ohiordr"
    booking_path = "OhioCampWeb"
    state_code = "OH"


class VirginiaStateParks(UseDirectProvider):
    """
    Virginia State Parks
    """

    base_url = "https://virginiardr.usedirect.com"
    campground_url = "https://VirginiaStateParks.com"
    rdr_path = "virginiardr"
    booking_path = "Web"
    state_code = "VA"


class ArizonaStateParks(UseDirectProvider):
    """
    Arizona State Parks
    """

    base_url = "https://azrdr.usedirect.com"
    campground_url = "https://ArizonaStateParks.com"
    rdr_path = "azrdr"
    booking_path = "reserve"
    state_code = "AZ"


class MissouriStateParks(UseDirectProvider):
    """
    Missouri State Parks
    """

    base_url = "https://msprdr.usedirect.com"
    campground_url = "https://icampmo1.usedirect.com"
    rdr_path = "msprdr"
    booking_path = "MSPWeb"
    state_code = "MO"


class MinnesotaStateParks(UseDirectProvider):
    """
    Minnesota State Parks
    """

    base_url = "https://mnrdr.usedirect.com"
    campground_url = "https://reservemn.usedirect.com"
    rdr_path = "minnesotardr"
    booking_path = "MinnesotaWeb"
    state_code = "MN"


# Regional / County Parks


class MaricopaCountyParks(UseDirectProvider):
    """
    Maricopa County Parks
    """

    base_url = "https://maricopardr.usedirect.com"
    campground_url = "https://www.maricopacountyparks.org"
    rdr_path = "maricopardr"
    booking_path = "MaricopaWeb/Facilities/SearchViewUnitAvailabity.aspx"
    booking_path_params = False
    state_code = "AZ"


class FairfaxCountyParks(UseDirectProvider):
    """
    Fairfax County Parks
    """

    base_url = "https://fairfaxrdr.usedirect.com"
    campground_url = "https://fairfax.usedirect.com"
    rdr_path = "FCPARDR"
    booking_path = "FairfaxFCPAWeb/Facilities/SearchViewUnitAvailabity.aspx"
    booking_path_params = False
    state_code = "VA"


class OregonMetro(UseDirectProvider):
    """
    Oregon Metro Parks
    """

    base_url = "https://oregonrdr.usedirect.com"
    campground_url = "https://reservemetro.usedirect.com"
    rdr_path = "oregonmetrordr"
    booking_path = "MetroWeb/Facilities/SearchViewUnitAvailabity.aspx"
    booking_path_params = False
    state_code = "OR"


# International Parks


class NorthernTerritory(UseDirectProvider):
    """
    Australian NorthernTerritory
    """

    base_url = "https://northernterritoryrdr.usedirect.com"
    campground_url = "https://parkbookings.nt.gov.au"
    rdr_path = "NorthernTerritoryRDR"
    booking_path = "Web/Facilities/SearchViewUnitAvailabity.aspx"
    booking_path_params = False
    state_code = "NT"



"""
Camply General Utilities
"""

import datetime
import logging
import sys
from datetime import date
from typing import Any, Callable, Iterable, List, Optional, Set, Tuple, Union

from camply import SearchWindow
from camply.containers.base_container import CamplyModel

logger = logging.getLogger(__name__)

ListLike = Union[List[Any], Set[Any], Tuple[Any]]


def is_list_like(obj: Any) -> bool:
    """
    Define if an object is list-like
    """
    return isinstance(obj, (list, set, tuple))


def make_list(obj, coerce: Optional[Callable] = None) -> Optional[List[Any]]:
    """
    Make Anything An Iterable Instance

    Parameters
    ----------
    obj: object
    coerce: Callable

    Returns
    -------
    List[object]
    """
    if obj is None:
        return None
    elif isinstance(obj, CamplyModel):
        return [coerce(obj) if coerce is not None else obj]
    elif is_list_like(obj) is True:
        if coerce is not None:
            return [coerce(item) for item in obj]
        else:
            return list(obj)
    else:
        return [coerce(obj) if coerce is not None else obj]


def handle_search_windows(
    start_date: Union[Iterable[str], str, Iterable[datetime.date], datetime.date],
    end_date: Union[Iterable[str], str, Iterable[datetime.date], datetime.date],
) -> Union[List[SearchWindow], SearchWindow]:
    """
    Handle Multiple Search Windows by the CLI
    """
    if isinstance(start_date, (str, date)):
        start_date = (start_date,)
        assert isinstance(end_date, (str, date))
        end_date = (end_date,)
    search_windows: List[SearchWindow] = []
    for field in [start_date, end_date]:
        if field is None or (isinstance(field, (tuple, list)) and len(field) == 0):
            logger.error("Campsite searches require a `start_date` and an `end_date`")
            sys.exit(1)
    if len(start_date) != len(end_date):
        logger.error(
            "When searching multiple date windows, you must provide the same amount "
            "of `--start-dates` as `--end-dates`"
        )
        sys.exit(1)
    for index, date_str in enumerate(start_date):
        search_windows.append(
            SearchWindow(start_date=date_str, end_date=end_date[index])
        )
    if len(search_windows) == 1:
        return search_windows[0]
    else:
        return search_windows


days_of_the_week_base = {
    "Monday": 0,
    "Tuesday": 1,
    "Wednesday": 2,
    "Thursday": 3,
    "Friday": 4,
    "Saturday": 5,
    "Sunday": 6,
}
days_of_the_week_mapping = days_of_the_week_base.copy()
days_of_the_week_mapping.update(
    {
        "MON": 0,
        "TUE": 1,
        "TUES": 1,
        "WED": 2,
        "THU": 3,
        "THUR": 3,
        "THURS": 3,
        "FRI": 4,
        "SAT": 5,
        "SUN": 6,
    }
)



"""
Yellowstone Testing Provider
"""

import logging
from datetime import datetime

import pytest

from camply.containers import AvailableCampsite, CampgroundFacility, SearchWindow
from camply.search import SearchRecreationDotGov
from tests.conftest import vcr_cassette

logger = logging.getLogger(__name__)


@pytest.fixture
def search_window() -> SearchWindow:
    """
    Get A RecDotGov Search Window For September 2023

    Returns
    -------
    SearchWindow
    """
    search_window = SearchWindow(
        start_date=datetime(2023, 9, 1),
        end_date=datetime(2023, 10, 1),
    )
    return search_window


@pytest.fixture
def recdotgov_recarea_finder(search_window) -> SearchRecreationDotGov:
    """
    Assemble The Searching Class

    Returns
    -------
    SearchYellowstone
    """
    recdotgov_finder = SearchRecreationDotGov(
        search_window=search_window, recreation_area=2584
    )  # Big Bend National Park
    logger.info("RecreationDotGov Recreation Area Searcher Established.")
    logger.info(f"Search Months: {recdotgov_finder.search_months}")
    return recdotgov_finder


@pytest.fixture
def recdotgov_campground_finder(search_window) -> SearchRecreationDotGov:
    """
    Assemble The Searching Class

    Returns
    -------
    SearchYellowstone
    """
    recdotgov_finder = SearchRecreationDotGov(
        search_window=search_window, campgrounds=234708
    )  # Apache Trout CampgroundApache Trout Campground
    logger.info("RecreationDotGov Campground Searcher Established.")
    logger.info(f"Search Months: {recdotgov_finder.search_months}")
    return recdotgov_finder


@pytest.fixture
def recdotgov_campsite_finder(search_window) -> SearchRecreationDotGov:
    """
    Assemble The Searching Class

    Returns
    -------
    SearchYellowstone
    """
    recdotgov_finder = SearchRecreationDotGov(
        search_window=search_window, campsites=93740
    )  # Site: O, Loop: Group Sites L-R - Chisos Basin Group Campground
    logger.info("RecreationDotGov Campground Searcher Established.")
    logger.info(f"Search Months: {recdotgov_finder.search_months}")
    return recdotgov_finder


@vcr_cassette
def test_get_searchable_campgrounds_recarea(
    recdotgov_recarea_finder,
) -> None:
    """
    Retrieve Campground Information for a Recreation Area
    """
    logger.info("Searching for Matching Recreation Area Campgrounds")
    campgrounds = recdotgov_recarea_finder._get_searchable_campgrounds()
    assert campgrounds
    for camp in campgrounds:
        assert isinstance(camp, CampgroundFacility)


@vcr_cassette
def test_get_searchable_campgrounds_campground(
    recdotgov_campground_finder,
) -> None:
    """
    Retrieve Campground Information for a Campground ID
    """
    logger.info("Searching for Matching Campgrounds to Campground ID")
    campgrounds = recdotgov_campground_finder._get_searchable_campgrounds()
    assert campgrounds
    for camp in campgrounds:
        assert isinstance(camp, CampgroundFacility)


@vcr_cassette
def test_get_all_campsites_recarea(recdotgov_recarea_finder) -> None:
    """
    Get all of the Yellowstone Campsites
    """
    logger.info("Searching for All Recreation Area Campsites")
    all_campsites = recdotgov_recarea_finder.get_all_campsites()
    SearchRecreationDotGov.assemble_availabilities(
        matching_data=all_campsites, log=True, verbose=False
    )
    assert all_campsites
    for camp in all_campsites:
        assert isinstance(camp, AvailableCampsite)


@vcr_cassette
def test_get_all_campsites_campground(
    recdotgov_campground_finder,
) -> None:
    """
    Get all of the Yellowstone Campsites
    """
    logger.info("Searching for Campground Specific Campsites")
    all_campsites = recdotgov_campground_finder.get_all_campsites()
    SearchRecreationDotGov.assemble_availabilities(
        matching_data=all_campsites, log=True, verbose=False
    )
    assert all_campsites
    for camp in all_campsites:
        assert isinstance(camp, AvailableCampsite)


@vcr_cassette
def test_get_campsite_specific_campgrounds(
    recdotgov_campsite_finder,
) -> None:
    """
    Get all of the Related Campgrounds
    """
    logger.info("Searching for Campground Specific Campsites")
    campgrounds = recdotgov_campsite_finder._get_campgrounds_by_campsite_id()
    assert campgrounds
    for camp in campgrounds:
        assert isinstance(camp, CampgroundFacility)


@vcr_cassette
def test_get_campsite_specific_results(
    recdotgov_campsite_finder,
) -> None:
    """
    Get the related Campsite's availability
    """
    logger.info("Searching for Campground Specific Campsites")
    all_campsites = recdotgov_campsite_finder.get_all_campsites()
    SearchRecreationDotGov.assemble_availabilities(
        matching_data=all_campsites, log=True, verbose=False
    )
    assert all_campsites
    for camp in all_campsites:
        assert isinstance(camp, AvailableCampsite)



"""
Test AlabamaStateParks Provider
"""

import os
import pathlib
from unittest import mock

from pytest import MonkeyPatch

from camply.providers import (
    AlabamaStateParks,
)
from tests.conftest import CamplyRunner, cli_status_checker, vcr_cassette


@vcr_cassette
@mock.patch.dict(os.environ, {"CAMPLY_PROVIDER": AlabamaStateParks.__name__})
def test_alabama_state_parks_recreation_areas(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    CLI Testing - AlabamaStateParks Recreation Areas
    """
    monkeypatch.setattr(AlabamaStateParks, "offline_cache_dir", tmp_path)
    test_command = """
    camply recreation-areas \
     --search "Park"
    """
    result = cli_runner.run_camply_command(test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert 'Using Camply Provider: "AlabamaStateParks"' in result.output
    assert "Blue Springs State Park" in result.output


@vcr_cassette
@mock.patch.dict(os.environ, {"CAMPLY_PROVIDER": AlabamaStateParks.__name__})
def test_alabama_state_parks_campgrounds(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    CLI Testing - AlabamaStateParks Campgrounds
    """
    monkeypatch.setattr(AlabamaStateParks, "offline_cache_dir", tmp_path)
    test_command = """
    camply campgrounds \
     --search "Park"
    """
    result = cli_runner.run_camply_command(test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert 'Using Camply Provider: "AlabamaStateParks"' in result.output
    assert "Wind Creek State Park" in result.output


@vcr_cassette
@mock.patch.dict(os.environ, {"CAMPLY_PROVIDER": AlabamaStateParks.__name__})
def test_alabama_state_parks_campsites(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    CLI Testing - AlabamaStateParks Campsites
    """
    monkeypatch.setattr(AlabamaStateParks, "offline_cache_dir", tmp_path)
    test_command = """
    camply campsites \
        --campground 114 \
        --start-date 2023-07-01 \
        --end-date 2023-07-14
    """
    result = cli_runner.run_camply_command(test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert 'Using Camply Provider: "AlabamaStateParks"' in result.output
    assert "total sites found in month of July" in result.output
    assert "Wind Creek State Park" in result.output



"""
UseDirect Provider
"""

import json
import logging
import pathlib
import sys
from abc import ABC, abstractmethod
from datetime import date, datetime, timedelta
from typing import Any, Dict, List, Optional, Union

import ratelimit
from fake_useragent import UserAgent
from pydantic import ValidationError

from camply.config import FileConfig
from camply.config.api_config import UseDirectConfig
from camply.containers import (
    AvailableCampsite,
    CampgroundFacility,
    CamplyModel,
    RecreationArea,
)
from camply.containers.data_containers import CampsiteLocation
from camply.containers.usedirect import (
    UseDirectAvailabilityResponse,
    UseDirectAvailabilitySlice,
    UseDirectAvailabilityUnit,
    UseDirectCityPark,
    UseDirectDetailedPlace,
    UseDirectFacilityMetadata,
    UseDirectMetadata,
)
from camply.exceptions import CamplyError
from camply.providers.base_provider import BaseProvider
from camply.utils.logging_utils import log_sorted_response

logger = logging.getLogger(__name__)


class UseDirectError(CamplyError):
    """
    UseDirect Provider Error
    """


class UseDirectProvider(BaseProvider, ABC):
    """
    Camply Provider for UseDirect RDR Campgrounds
    """

    usedirect_city_parks: Dict[int, UseDirectCityPark] = {}
    usedirect_rec_areas: Dict[int, RecreationArea] = {}
    usedirect_campgrounds: Dict[int, CampgroundFacility] = {}
    usedirect_unit_categories: Dict[int, str] = {}
    usedirect_unit_type_groups: Dict[int, str] = {}
    usedirect_campsites: Dict[int, UseDirectAvailabilityUnit] = {}
    campsite_ids: List[int] = []
    metadata_refreshed: bool = False
    active_search: bool = False

    __offline_cache_dir__: Optional[pathlib.Path] = None

    rdr_path: str = "rdr"

    booking_path_params: bool = True
    booking_path: str = "Web/Default.aspx"

    @property
    @abstractmethod
    def base_url(self) -> str:
        """
        Base URL for the Provider
        """
        pass

    @property
    @abstractmethod
    def campground_url(self) -> str:
        """
        Campground URL for the Provider
        """
        pass

    @property
    @abstractmethod
    def state_code(self) -> str:
        """
        State Code for the Provider
        """
        pass

    @property
    def offline_cache_dir(self) -> pathlib.Path:
        """
        Offline Cache Directory
        """
        if self.__offline_cache_dir__ is None:
            return FileConfig.USEDIRECT_PROVIDER / self.__class__.__name__
        else:
            return self.__offline_cache_dir__

    def refresh_metadata(self) -> None:
        """
        Refresh All the Campground Metadata

        This is the way that this provider caches all of its metadata
        offline. It makes a number of GET requests and saves the entire output
        as JSON alongside the provider code itself (<subdomain>.usedirect.com):

        - /rdr/rdr/search/filters
        - /rdr/rdr/search/citypark
        - /rdr/rdr/search/places
        - /rdr/rdr/search/facilities

        Returns
        -------
        None
        """
        if self.metadata_refreshed is False:
            self._get_campground_metadata()
            self._get_city_parks()
            self._get_places()
            self._get_facilities()
        self.metadata_refreshed = True

    def search_for_recreation_areas(
        self,
        query: Optional[str] = None,
        state: Optional[str] = None,
    ) -> List[RecreationArea]:
        """
        Retrieve Recreation Areas

        Parameters
        ----------
        state: str
        query: str

        Returns
        -------
        List[RecreationArea]
        """
        if state is not None and state.upper() != self.state_code.upper():
            raise CamplyError(
                f"{self.__class__.__name__} doesn't support states outside {self.state_code}"
            )
        if query is None:
            logger.error(
                "You must provide a search string to search `UseDirect` Recreation Areas"
            )
            sys.exit(1)
        logger.info(f'Searching for Recreation Areas: "{query}"')
        self.refresh_metadata()
        found_recareas = [
            rec_area
            for rec_area in self.usedirect_rec_areas.values()
            if self._search_camply_model(query=query, model=rec_area) is True
        ]
        return found_recareas

    def find_campgrounds(
        self,
        search_string: Optional[str] = None,
        rec_area_id: Optional[List[int]] = None,
        state: Optional[str] = None,
        verbose: bool = True,
        campground_id: Optional[List[int]] = None,
        **kwargs: Any,
    ) -> List[CampgroundFacility]:
        """
        Search A Facility via Offline Metadata

        Parameters
        ----------
        rec_area_id: Optional[int]
        state: Optional[str]
        search_string: Optional[str]
        campground_id: Optional[List[int]]

        Returns
        -------
        List[CampgroundFacility]
        """
        self.active_search = True
        if state is not None and state.upper() != self.state_code:
            raise CamplyError(
                f"{self.__class__.__name__} doesn't support states outside {self.state_code}"
            )
        if campground_id is None:
            campground_id = []
        if rec_area_id is None:
            rec_area_id = []
        if all([rec_area_id == [], search_string is None, campground_id == []]):
            logger.error(
                "You must provide a search string, campground ID, or recreation area ID "
                "to search on UseDirect"
            )
            sys.exit(1)
        if search_string is not None:
            logger.info(f'Searching for Campgrounds: "{search_string}"')
        self.refresh_metadata()
        found_campgrounds = self._search_for_campgrounds(
            campground_id=campground_id,
            rec_area_id=rec_area_id,
            search_string=search_string,
        )
        if verbose is True:
            logger.info(f"{len(found_campgrounds)} Matching Campgrounds Found")
            log_sorted_response(found_campgrounds)
        self.active_search = False
        return found_campgrounds

    def _search_for_campgrounds(
        self,
        campground_id: List[int],
        rec_area_id: List[int],
        search_string: Optional[str],
    ) -> List[CampgroundFacility]:
        """
        Filter a campground array

        Parameters
        ----------
        campground_id: List[int]
        rec_area_id: List[int]
        search_string: Optional[str]

        Returns
        -------
        List[CampgroundFacility]
        """
        found_campgrounds: List[CampgroundFacility] = []
        if len(campground_id) >= 1:
            for camp_id in campground_id:
                found_campgrounds += [
                    campground
                    for campground in self.usedirect_campgrounds.values()
                    if campground.facility_id == int(camp_id)
                ]
        elif len(rec_area_id) >= 1:
            for rec_area in rec_area_id:
                found_campgrounds += [
                    campground
                    for campground in self.usedirect_campgrounds.values()
                    if campground.recreation_area_id == int(rec_area)
                ]
        else:
            assert isinstance(search_string, str)
            found_campgrounds = [
                campground
                for campground in self.usedirect_campgrounds.values()
                if self._search_camply_model(query=search_string, model=campground)
                is True
            ]
        return found_campgrounds

    @ratelimit.sleep_and_retry
    @ratelimit.limits(calls=1, period=1)
    def get_campsites_response(
        self,
        campground_id: int,
        start_date: Union[datetime, date],
        end_date: Union[datetime, date],
        is_ada: Optional[bool] = None,
        min_vehicle_length: Optional[int] = None,
        unit_category_id: Optional[int] = None,
        web_only: Optional[bool] = True,
        unit_type_group_ids: Optional[List[int]] = None,
        sleeping_unit_id: Optional[int] = None,
        unit_sort: Optional[str] = "orderby",
        in_season_only: Optional[bool] = True,
    ) -> UseDirectAvailabilityResponse:
        """
        Get Campsites from UseDirect

        Parameters
        ----------
        campground_id: int
            Facility ID of the campground
        start_date: Union[datetime, date]
            Search Start Date
        end_date: Union[datetime, date]
            Search End Date
        is_ada: Optional[bool]
            Search for ADA sites
        min_vehicle_length: Optional[int]
            Minimum Vehicle Length - defaults to 0 which doesn't filter
        unit_category_id: Optional[int]
            Unit Category ID (typically 0)
        web_only: Optional[bool]
            Search for sights bookable online
        unit_type_group_ids: Optional[List[int]]
            UnitTypeGroupIds - Search Param
        sleeping_unit_id: Optional[int]
            SleepingUnitId - search param
        unit_sort: Optional[str]
            Sort Order
        in_season_only: Optional[bool]
            Searching for in-season only campgrounds

        Returns
        -------
        UseDirectAvailabilityResponse
        """
        data = {
            "IsADA": is_ada,
            "MinVehicleLength": min_vehicle_length,
            "UnitCategoryId": unit_category_id,
            "StartDate": start_date.strftime(UseDirectConfig.DATE_FORMAT),
            "WebOnly": web_only,
            "UnitTypesGroupIds": []
            if unit_type_group_ids is None
            else unit_type_group_ids,
            "SleepingUnitId": sleeping_unit_id,
            "EndDate": end_date.strftime(UseDirectConfig.DATE_FORMAT),
            "UnitSort": unit_sort,
            "InSeasonOnly": in_season_only,
            "FacilityId": campground_id,
        }
        non_null_data = {
            key: value for key, value in data.items() if value not in [None, [], ""]
        }
        url = f"{self.base_url}/{self.rdr_path}/{UseDirectConfig.AVAILABILITY_ENDPOINT}"
        random_ua = UserAgent(browsers=["chrome"]).random
        self.json_headers["User-Agent"] = random_ua
        response = self.make_http_request_retry(
            url=url,
            method="POST",
            data=json.dumps(non_null_data),
            headers=self.json_headers,
        )
        response_json = response.json()
        try:
            return UseDirectAvailabilityResponse(**response_json)
        except ValidationError as e:
            raise
            error_message = (
                "Error Parsing UseDirect Availability Response "
                f"- Facility ID # {campground_id}."
            )
            if "Message" in response_json:
                error_message += " " + response_json["Message"]
            raise UseDirectError(error_message) from e

    def get_campsites(
        self,
        campground_id: int,
        start_date: Union[datetime, date],
        end_date: Union[datetime, date],
        is_ada: Optional[bool] = None,
        min_vehicle_length: Optional[int] = None,
        unit_category_id: Optional[int] = None,
        web_only: Optional[bool] = True,
        unit_type_group_ids: Optional[List[int]] = None,
        sleeping_unit_id: Optional[int] = None,
        unit_sort: Optional[str] = "orderby",
        in_season_only: Optional[bool] = True,
    ) -> List[AvailableCampsite]:
        """
        Get Campsites from UseDirect

        Parameters
        ----------
        campground_id: int
            Facility ID of the campground
        start_date: Union[datetime, date]
            Search Start Date
        end_date: Union[datetime, date]
            Search End Date
        is_ada: Optional[bool]
            Search for ADA sites
        min_vehicle_length: Optional[int]
            Minimum Vehicle Length - defaults to 0 which doesn't filter
        unit_category_id: Optional[int]
            Unit Category ID (typically 0)
        web_only: Optional[bool]
            Search for sights bookable online
        unit_type_group_ids: Optional[List[int]]
            UnitTypeGroupIds - Search Param
        sleeping_unit_id: Optional[int]
            SleepingUnitId - search param
        unit_sort: Optional[str]
            Sort Order
        in_season_only: Optional[bool]
            Searching for in-season only campgrounds

        Returns
        -------
        List[AvailableCampsite]
        """
        self.refresh_metadata()
        availability_response = self.get_campsites_response(
            campground_id=campground_id,
            start_date=start_date,
            end_date=end_date,
            is_ada=is_ada,
            min_vehicle_length=min_vehicle_length,
            unit_category_id=unit_category_id,
            web_only=web_only,
            unit_type_group_ids=unit_type_group_ids,
            sleeping_unit_id=sleeping_unit_id,
            unit_sort=unit_sort,
            in_season_only=in_season_only,
        )
        campsites: List[AvailableCampsite] = []
        if availability_response.Facility.Units is None:
            return campsites
        for _campground_unit_id, unit in availability_response.Facility.Units.items():
            for _slice_date, availability_slice in unit.Slices.items():
                campsite = self._get_available_campsite(
                    availability_slice=availability_slice,
                    availability_response=availability_response,
                    unit=unit,
                )
                campsite_available = campsite.availability_status == "Available"
                if campsite_available is True:
                    if (
                        len(self.campsite_ids) == 0
                        or campsite.campsite_id in self.campsite_ids
                    ):
                        campsites.append(campsite)
        return campsites

    def _get_available_campsite(
        self,
        availability_slice: UseDirectAvailabilitySlice,
        availability_response: UseDirectAvailabilityResponse,
        unit: UseDirectAvailabilityUnit,
    ) -> AvailableCampsite:
        """
        Create an AvailableCampsite Object from the Availability Grid Response

        Parameters
        ----------
        availability_slice: UseDirectAvailabilitySlice
        availability_response: UseDirectAvailabilityResponse
        unit: UseDirectAvailabilityUnit

        Returns
        -------
        AvailableCampsite
        """
        start_date = datetime.fromordinal(availability_slice.Date.toordinal())
        facility_id = availability_response.Facility.FacilityId
        facility = self.usedirect_campgrounds[facility_id]
        recreation_area = self.usedirect_rec_areas[facility.recreation_area_id]
        booking_url = f"{self.campground_url}/{self.booking_path}"
        if self.booking_path_params is True:
            booking_url = f"{booking_url}#!park/{recreation_area.recreation_area_id}/{facility_id}"
        if unit.UnitCategoryId is None:
            unit.UnitCategoryId = -1
        if unit.UnitTypeGroupId is None:
            unit.UnitTypeGroupId = -1
        campsite_type = self.usedirect_unit_categories.get(unit.UnitCategoryId, None)
        campsite_use_type = self.usedirect_unit_type_groups.get(
            unit.UnitTypeGroupId, None
        )
        campsite = AvailableCampsite(
            campsite_id=unit.UnitId,
            booking_date=start_date,
            booking_end_date=start_date + timedelta(days=1),
            booking_nights=1,
            campsite_site_name=unit.Name,
            availability_status=(
                "Available" if availability_slice.IsFree is True else "Unavailable"
            ),
            recreation_area=recreation_area.recreation_area,
            recreation_area_id=facility.recreation_area_id,
            facility_name=facility.facility_name,
            facility_id=facility.facility_id,
            booking_url=booking_url,
            campsite_occupancy=(0, 1),
            campsite_type=campsite_type,
            campsite_use_type=campsite_use_type,
            location=CampsiteLocation(
                latitude=availability_response.Facility.Latitude,
                longitude=availability_response.Facility.Longitude,
            ),
        )
        return campsite

    def _fetch_metadata_from_disk(
        self, file_path: pathlib.Path
    ) -> Optional[Union[Dict[Any, Any], List[Dict[Any, Any]]]]:
        """
        Cache Metadata Locally and Invalidate after a day

        Parameters
        ----------
        file_path: pathlib.Path

        Returns
        -------
        Optional[Dict[Any, Any]]
        """
        if file_path.exists() is False:
            data = None
        else:
            modified_time = datetime.utcfromtimestamp(file_path.stat().st_mtime)
            current_time = datetime.utcnow()
            if (
                current_time - modified_time > timedelta(days=1)
                and self.active_search is False
            ):
                data = None
            else:
                json_body: Dict[Any, Any] = json.loads(
                    file_path.read_text(encoding="utf-8")
                )
                data = json_body
        if data is None:
            logger.debug("Refreshing UseDirect Metadata: %s", file_path.name)
        return data

    def _get_campground_metadata(self) -> UseDirectMetadata:
        """
        Return Metadata for Campgrounds

        Returns
        -------
        UseDirectMetadata
        """
        metadata_file = self.offline_cache_dir.joinpath("filters.json")
        campground_metadata = self._fetch_metadata_from_disk(file_path=metadata_file)
        if campground_metadata is None:
            self.offline_cache_dir.mkdir(parents=True, exist_ok=True)
            url = f"{self.base_url}/{self.rdr_path}/{UseDirectConfig.METADATA_PREFIX}"
            resp = self.make_http_request_retry(url=url)
            resp.raise_for_status()
            campground_metadata = resp.json()
            metadata_file.write_text(json.dumps(campground_metadata, indent=2))
        data = UseDirectMetadata(**campground_metadata)
        self.usedirect_unit_categories = {
            item.UnitCategoryId: item.UnitCategoryName for item in data.UnitCategories
        }
        self.usedirect_unit_type_groups = {
            item.UnitTypesGroupId: item.UnitTypesGroupName
            for item in data.UnitTypesGroups
        }
        return data

    def _get_city_parks(self) -> Dict[int, UseDirectCityPark]:
        """
        Fetch Metadata On Every CityPark

        Returns
        -------
        Dict[int, UseDirectCityPark]
        """
        metadata_file = self.offline_cache_dir.joinpath("cityparks.json")
        city_park_data = self._fetch_metadata_from_disk(file_path=metadata_file)
        if city_park_data is None:
            url = f"{self.base_url}/{self.rdr_path}/{UseDirectConfig.CITYPARK_ENDPOINT}"
            resp = self.make_http_request_retry(url=url)
            resp.raise_for_status()
            city_park_data: Dict[str, Dict[str, Any]] = resp.json()
            metadata_file.write_text(json.dumps(city_park_data, indent=2))
        self.usedirect_city_parks: Dict[int, UseDirectCityPark] = {
            int(city_park_id): UseDirectCityPark(**city_park_json)
            for city_park_id, city_park_json in city_park_data.items()
            if city_park_json["Name"] is not None
        }
        return self.usedirect_city_parks

    def _get_places(self) -> Dict[int, UseDirectDetailedPlace]:
        """
        Fetch Metadata On Every Place

        Returns
        -------
        Dict[int, UseDirectDetailedPlace]
        """
        metadata_file = self.offline_cache_dir.joinpath("places.json")
        places_data = self._fetch_metadata_from_disk(file_path=metadata_file)
        if places_data is None:
            url = f"{self.base_url}/{self.rdr_path}/{UseDirectConfig.LIST_PLACES_ENDPOINT}"
            resp = self.make_http_request_retry(url=url)
            resp.raise_for_status()
            places_data: List[Dict[str, Any]] = resp.json()
            metadata_file.write_text(json.dumps(places_data, indent=2))
        places_validated = [
            UseDirectDetailedPlace(**place_json) for place_json in places_data
        ]
        places_data_validated: Dict[int, UseDirectDetailedPlace] = {
            item.PlaceId: item for item in places_validated
        }
        self.usedirect_rec_areas: Dict[int, RecreationArea] = {
            place.PlaceId: RecreationArea(
                recreation_area=place.Name,
                recreation_area_id=place.PlaceId,
                recreation_area_location=f"{place.City.title()}, {place.State}",
                description=place.Description,
            )
            for place in places_data_validated.values()
        }
        return places_data_validated

    def _get_facilities(self) -> Dict[int, UseDirectFacilityMetadata]:
        """
        Fetch Metadata On Every Facility

        Returns
        -------
        Dict[int, UseDirectFacilityMetadata]
        """
        metadata_file = self.offline_cache_dir.joinpath("facilities.json")
        facilities_data = self._fetch_metadata_from_disk(file_path=metadata_file)
        if facilities_data is None:
            url = f"{self.base_url}/{self.rdr_path}/{UseDirectConfig.LIST_FACILITIES_ENDPOINT}"
            resp = self.make_http_request_retry(url=url)
            resp.raise_for_status()
            facilities_data: List[Dict[str, Any]] = resp.json()
            metadata_file.write_text(json.dumps(facilities_data, indent=2))
        if not isinstance(facilities_data, list):
            raise CamplyError("Unexpected data from %s", metadata_file)
        facilities_validated = [
            UseDirectFacilityMetadata(**facility_json)
            for facility_json in facilities_data
        ]
        facilities_data_validated: Dict[int, UseDirectFacilityMetadata] = {
            item.FacilityId: item for item in facilities_validated
        }
        self.usedirect_campgrounds: Dict[int, CampgroundFacility] = {}
        for facility in facilities_data_validated.values():
            rec_area = self.usedirect_rec_areas.get(facility.PlaceId, None)
            if rec_area is not None:
                self.usedirect_campgrounds[facility.FacilityId] = CampgroundFacility(
                    facility_name=facility.Name,
                    facility_id=facility.FacilityId,
                    recreation_area_id=facility.PlaceId,
                    recreation_area=rec_area.recreation_area,
                )
        return facilities_data_validated

    @classmethod
    def _search_camply_model(cls, query: str, model: CamplyModel) -> bool:
        """
        Search a Camply Model

        Parameters
        ----------
        query: str
        model: CamplyModel

        Returns
        -------
        bool
        """
        return any(
            query.lower() in str(value).lower() for value in model.dict().values()
        )

    def get_campsites_per_facility(
        self, facility_id: int
    ) -> List[UseDirectAvailabilityUnit]:
        """
        Get Campsites Per Facility

        Parameters
        ----------
        facility_id: int

        Returns
        -------
        List[UseDirectAvailabilityUnit]
        """
        resp = self.get_campsites_response(
            campground_id=facility_id, start_date=date.today(), end_date=date.today()
        )
        campsites: List[UseDirectAvailabilityUnit] = list(resp.Facility.Units.values())
        for campsite in campsites:
            campsite.FacilityId = facility_id
        return campsites

    def get_campsite_metadata(
        self, facility_ids: List[int]
    ) -> Dict[int, UseDirectAvailabilityUnit]:
        """
        Get the Campsite Metadata

        Parameters
        ----------
        facility_ids: List[int]
            List of facility ids to fetch metadata for

        Returns
        -------
        Dict[int, UseDirectAvailabilityUnit]
        """
        campsites: Dict[int, UseDirectAvailabilityUnit] = {}
        for facility_id in facility_ids:
            found_campsites = self.get_campsites_per_facility(facility_id=facility_id)
            campsite_dict = {item.UnitId: item for item in found_campsites}
            campsites.update(campsite_dict)
        self.usedirect_campsites.update(campsites)
        return campsites

    def _prepare_facility_ids(
        self,
        recreation_area_ids: Optional[List[int]] = None,
        campground_ids: Optional[List[int]] = None,
    ) -> List[int]:
        """
        Prepare Facility Ids

        Parameters
        ----------
        recreation_area_ids: Optional[List[int]]
        campground_ids: Optional[List[int]]

        Returns
        -------
        List[int]
        """
        if not self.usedirect_campgrounds:
            self.refresh_metadata()
        recreation_area_ids = recreation_area_ids or []
        campground_ids = campground_ids or []
        facility_ids = []
        if len(recreation_area_ids) == 0 and len(campground_ids) == 0:
            raise CamplyError("Must specify either a recreation area or campground id")
        elif len(recreation_area_ids) > 0:
            logger.info(
                "Searching %s Recreation Areas for campgrounds",
                len(recreation_area_ids),
            )
            for recreation_area_id in recreation_area_ids:
                facility_ids += [
                    facility_id
                    for facility_id, facility in self.usedirect_campgrounds.items()
                    if facility.recreation_area_id == int(recreation_area_id)
                ]
        else:
            facility_ids = campground_ids
        facility_ids = [int(x) for x in facility_ids]
        return facility_ids

    def validate_campsites(self, campsites: List[int], facility_ids: List[int]) -> None:
        """
        Validate Campsites

        Parameters
        ----------
        campsites: List[int]

        Returns
        -------
        None
        """
        self.get_campsite_metadata(facility_ids=facility_ids)
        for campsite_id in campsites:
            if campsite_id not in self.usedirect_campsites:
                raise CamplyError(f"Campsite {campsite_id} not found")
            else:
                campsite = self.usedirect_campsites[campsite_id]
                logger.info("Searching Specific campsite: %s", campsite.Name)
        self.campsite_ids = list(campsites)



"""
Yellowstone Lodging Web Searching Utilities
"""

import logging
from datetime import datetime, timedelta
from typing import Any, List, Optional, Set, Union

import pandas as pd

from camply.config.api_config import YellowstoneConfig
from camply.containers import AvailableCampsite, RecreationArea, SearchWindow
from camply.exceptions import SearchError
from camply.providers import Yellowstone
from camply.search.base_search import BaseCampingSearch
from camply.utils import make_list
from camply.utils.logging_utils import log_sorted_response

logger = logging.getLogger(__name__)


class SearchYellowstone(BaseCampingSearch):
    """
    Searches on YellowstoneNationalParkLodges.com for Campsites
    """

    recreation_area = Yellowstone.recreation_area
    provider_class = Yellowstone
    list_campsites_supported: bool = False

    # noinspection PyUnusedLocal
    def __init__(
        self,
        search_window: Union[SearchWindow, List[SearchWindow]],
        weekends_only: bool = False,
        campgrounds: Optional[Union[List[str], str]] = None,
        nights: int = 1,
        offline_search: bool = False,
        offline_search_path: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Initialize with Search Parameters

        Parameters
        ----------
        search_window: Union[SearchWindow, List[SearchWindow]]
            Search Window tuple containing start date and End Date
        weekends_only: bool
            Whether to only search for Camping availabilities on the weekends (Friday /
            Saturday nights)
        campgrounds: Optional[Union[List[str], str]]
            Campground ID or List of Campground IDs
        nights: int
            minimum number of consecutive nights to search per campsite,defaults to 1
        offline_search: bool
            When set to True, the campsite search will both save the results of the
            campsites it's found, but also load those campsites before beginning a
            search for other campsites.
        offline_search_path: Optional[str]
            When offline search is set to True, this is the name of the file to be saved/loaded.
            When not specified, the filename will default to `camply_campsites.json`
        """
        super().__init__(
            search_window=search_window,
            weekends_only=weekends_only,
            nights=nights,
            offline_search=offline_search,
            offline_search_path=offline_search_path,
            **kwargs,
        )
        self.campgrounds = make_list(campgrounds)

    def get_all_campsites(self) -> List[AvailableCampsite]:
        """
        Search for all matching campsites in Yellowstone.

        Returns
        -------
        List[AvailableCampsite]
        """
        all_campsites = []
        searchable_campgrounds = self._get_searchable_campgrounds()
        this_month = datetime.now().date().replace(day=1)
        for month in self.search_months:
            if month >= this_month:
                all_campsites += self.campsite_finder.get_monthly_campsites(
                    month=month, nights=None if self.nights == 1 else self.nights
                )
        matching_campsites = self._filter_campsites_to_campgrounds(
            campsites=all_campsites, searchable_campgrounds=searchable_campgrounds
        )
        campsite_df = self.campsites_to_df(campsites=matching_campsites)
        campsite_df_validated = self._filter_date_overlap(campsites=campsite_df)
        time_window_end = max(self.search_days) + timedelta(days=1)
        compiled_campsite_df = campsite_df_validated[
            campsite_df_validated.booking_end_date <= pd.Timestamp(time_window_end)
        ]
        compiled_campsites = self.df_to_campsites(campsite_df=compiled_campsite_df)
        return compiled_campsites

    def _get_searchable_campgrounds(self) -> Optional[Set[str]]:
        """
        Return the Campgrounds for the Camping Search

        Returns
        -------
        Optional[Set[str]]
        """
        if self.campgrounds in [None, []]:
            return None
        supported_campsites = set(YellowstoneConfig.YELLOWSTONE_CAMPGROUNDS.keys())
        selected_campsites = set(self.campgrounds)
        searchable_campgrounds = supported_campsites.intersection(selected_campsites)
        if len(searchable_campgrounds) == 0:
            campground_ids = [
                f"`{key}` ({value})"
                for key, value in YellowstoneConfig.YELLOWSTONE_CAMPGROUNDS.items()
            ]
            error_message = (
                "You must supply a YellowstoneNationalParkLodges supported "
                "campground ID. Current supported Campground IDs: "
                f"{', '.join(campground_ids)}"
            )
            logger.error(error_message)
            raise SearchError(error_message)
        logger.info(f"{len(searchable_campgrounds)} Matching Campgrounds Found")
        for campground in searchable_campgrounds:
            logger.info(
                f"⛰  {YellowstoneConfig.YELLOWSTONE_RECREATION_AREA_FORMAL_NAME} "
                f"(#{YellowstoneConfig.YELLOWSTONE_RECREATION_AREA_ID}) - 🏕  "
                f"{YellowstoneConfig.YELLOWSTONE_CAMPGROUNDS[campground]} ({campground})"
            )
        return searchable_campgrounds

    def _filter_campsites_to_campgrounds(
        self, campsites: List[AvailableCampsite], searchable_campgrounds: Set[str]
    ) -> List[AvailableCampsite]:
        """
        Filter Campsites Down to Matching Campgrounds

        Parameters
        ----------
        campsites: List[AvailableCampsite]
        searchable_campgrounds: Set[str]

        Returns
        -------
        List[AvailableCampsite]
        """
        if self.campgrounds in [None, []]:
            return campsites
        matching_campsites = [
            campsite
            for campsite in campsites
            if campsite.facility_id in searchable_campgrounds
        ]
        return matching_campsites

    @classmethod
    def find_recreation_areas(cls, **kwargs) -> List[RecreationArea]:
        """
        Return the Yellowstone Recreation Area Object
        """
        log_sorted_response([cls.recreation_area])
        return [cls.recreation_area]

    def list_campsite_units(self) -> Any:
        """
        List Campsite Units

        Returns
        -------
        Any
        """
        raise NotImplementedError



"""
Going to Camp Providers
"""

from camply.containers import RecreationArea

# Map going to camp subdomains to RecreationAreas
RECREATION_AREAS = {
    "longpoint.goingtocamp.com": RecreationArea(
        recreation_area="Long Point Region",
        recreation_area_id=1,
        recreation_area_location="Ontario, CA",
    ),
    "stclair.goingtocamp.com": RecreationArea(
        recreation_area="St. Clair Region",
        recreation_area_id=2,
        recreation_area_location="Ontario, CA",
    ),
    "washington.goingtocamp.com": RecreationArea(
        recreation_area="Washington State Parks",
        recreation_area_id=3,
        recreation_area_location="Washington, USA",
    ),
    "maitlandvalley.goingtocamp.com": RecreationArea(
        recreation_area="Maitland Valley",
        recreation_area_id=4,
        recreation_area_location="Ontario, CA",
    ),
    "saugeen.goingtocamp.com": RecreationArea(
        recreation_area="Saugeen Valley",
        recreation_area_id=5,
        recreation_area_location="Ontario, CA",
    ),
    "tacomapower.goingtocamp.com": RecreationArea(
        recreation_area="Tacoma Power Parks",
        recreation_area_id=6,
        recreation_area_location="Washington, USA",
    ),
    "wisconsin.goingtocamp.com": RecreationArea(
        recreation_area="Wisconsin State Parks",
        recreation_area_id=7,
        recreation_area_location="Wisconsin, USA",
    ),
    "ahtrails.ca": RecreationArea(
        recreation_area="Algonquin Highlands",
        recreation_area_id=8,
        recreation_area_location="Ontario, CA",
    ),
    "parkreservations.maryland.gov": RecreationArea(
        recreation_area="Maryland State Parks",
        recreation_area_id=9,
        recreation_area_location="Maryland, USA",
    ),
    "reservations.ncc-ccn.gc.ca": RecreationArea(
        recreation_area="Gatineau Park",
        recreation_area_id=10,
        recreation_area_location="Ottawa-Gatineau, Ontario-Quebec, CA",
    ),
    "nlcamping.ca": RecreationArea(
        recreation_area="Newfoundland & Labrador Provincial Parks",
        recreation_area_id=11,
        recreation_area_location="Newfoundland and Labrador, CA",
    ),
    "camping.bcparks.ca": RecreationArea(
        recreation_area="BC Parks",
        recreation_area_id=12,
        recreation_area_location="British Columbia, CA",
    ),
    "novascotia.goingtocamp.com": RecreationArea(
        recreation_area="Nova Scotia Parks",
        recreation_area_id=13,
        recreation_area_location="Nova Scotia, CA",
    ),
    "reservation.pc.gc.ca": RecreationArea(
        recreation_area="Parks Canada",
        recreation_area_id=14,
        recreation_area_location="CA",
    ),
    "manitoba.goingtocamp.com": RecreationArea(
        recreation_area="Manitoba Parks",
        recreation_area_id=15,
        recreation_area_location="Manitoba, CA",
    ),
    "parcsnbparks.ca": RecreationArea(
        recreation_area="New Brunswick Provincial Parks",
        recreation_area_id=16,
        recreation_area_location="New Brunswick, CA",
    ),
    "midnrreservations.com": RecreationArea(
        recreation_area="Michigan State Parks",
        recreation_area_id=17,
        recreation_area_location="Michigan, USA",
    ),
}



"""
Pytest Fixtures Shared Across all Unit Tests
"""

import datetime
import logging
from textwrap import dedent
from typing import Any, Dict

import pytest
from click.testing import CliRunner, Result
from freezegun import freeze_time

from camply import AvailableCampsite
from camply.cli import camply_command_line

logger = logging.getLogger(__name__)
[
    logging.getLogger(loggo).setLevel(logging.WARNING)
    for loggo in [
        "vcr.cassette",
        "vcr.request",
        "vcr.matchers",
    ]
]
module_scope = pytest.fixture(scope="module")


@pytest.fixture(autouse=True)
def freeze_current_time():
    """
    Freeze the Current Time to April 28, 2023 at Noon

    Since camply saves the responses of the API calls, we need to freeze the time
    to ensure the responses are the same across all tests.
    """
    year = 2023
    time_of_year = [4, 28, 12, 0, 0]  # April 28th
    frozen_time = datetime.datetime(year, *time_of_year)
    with freeze_time(frozen_time, tick=True):
        yield


class CamplyRunner(CliRunner):
    """
    Custom CLI Runner for Camply
    """

    def run_camply_command(self, command: str) -> Result:
        """
        Run a Camply Command and Return the Result

        Parameters
        ----------
        command

        Returns
        -------
        Result
        """
        parsed_command = self.parse_camply_command(command=command)
        logger.debug("Camply CLI: %s", parsed_command)
        result = self.invoke(cli=camply_command_line, args=parsed_command)
        return result

    @classmethod
    def parse_camply_command(cls, command: str) -> str:
        """
        Parse a Camply CLI Command to a Parseable Str

        Parameters
        ----------
        command: str

        Returns
        -------
        str
        """
        command_parsed = dedent(command).strip()
        for r in (("\\", ""), ("\n", ""), ("\t", " "), ("  ", " "), ("camply ", "")):
            command_parsed = command_parsed.replace(*r)
        return command_parsed


@pytest.fixture
def cli_runner() -> CamplyRunner:
    """
    Fixture for invoking command-line interfaces.
    """
    return CamplyRunner()


def cli_status_checker(result: Result, exit_code_zero: bool = True) -> None:
    """
    Handle Exceptions from the CLI

    Parameters
    ----------
    result : Result
        CliRunner Invoke Result
    exit_code_zero: bool
        Whether the exit code should be `0` - defaults to True
    """
    try:
        assert (result.exit_code == 0) == exit_code_zero
    except AssertionError as e:
        logger.exception(result.exception, exc_info=result.exc_info)
        raise result.exception from e


@module_scope
def vcr_config() -> Dict[str, Any]:
    """
    VCR Cassette Privacy Enforcer

    This fixture ensures the API Credentials are obfuscated

    Returns
    -------
    Dict[str, Any]
    """
    return {
        "filter_headers": [("authorization", "REDACTED"), ("apikey", "REDACTED")],
        "filter_query_parameters": [("user", "REDACTED"), ("token", "REDACTED")],
    }


# Decorator Object to Use pyvcr Cassettes on Unit Tests (see `pytest-vcr`)
# pass `--vcr-record=none` to pytest CI runs to ensure new cassettes are generated
vcr_cassette = pytest.mark.vcr(scope="module")


@pytest.fixture
def available_campsite() -> AvailableCampsite:
    """
    Test AvailableCampsite Instance
    """
    return AvailableCampsite(
        campsite_id=100,
        booking_date=datetime.datetime(2023, 9, 1),
        booking_end_date=datetime.datetime(2023, 9, 2),
        booking_nights=1,
        campsite_site_name="Test Campsite Name",
        campsite_loop_name="A1",
        campsite_type="Test",
        campsite_occupancy=(1, 5),
        campsite_use_type="Test",
        availability_status="Available",
        recreation_area="Test Recreation Area",
        recreation_area_id=20,
        facility_name="Test Campground",
        facility_id=50,
        booking_url="https://youtu.be/eBGIQ7ZuuiU",
        permitted_equipment=[],
        campsite_attributes=[],
    )



"""
BaseProvider Base Class
"""

import logging
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional, Type, Union

import requests
import tenacity
from fake_useragent import UserAgent

from camply.config import SearchConfig
from camply.config.api_config import APIConfig
from camply.containers import CampgroundFacility

logger = logging.getLogger(__name__)


class ProviderError(Exception):
    """
    General Provider Error
    """


class ProviderSearchError(ProviderError):
    """
    Searching Error
    """


class BaseProvider(ABC):
    """
    Base Provider Class
    """

    RETRY_CONFIG: Type[APIConfig] = APIConfig
    FIVE_HUNDRED_STATUS_CODES = [
        # Official Server Errors
        500,  # Internal Server Error
        501,  # Not Implemented
        502,  # Bad Gateway
        503,  # Service Unavailable
        504,  # Gateway Timeout
        505,  # HTTP Version Not Supported
        506,  # Variant Also Negotiates
        507,  # Insufficient Storage
        508,  # Loop Detected
        510,  # Not Extended
        511,  # Network Authentication Required
        # Unofficial Server Errors
        509,  # Bandwidth Limit Exceeded
        529,  # Site is overloaded
        530,  # Site is frozen
        598,  # Network read timeout error
        599,  # Network connect timeout error
        # Vendor Errors
        520,  # Unknown Error
        521,  # Web Server Is Down
        522,  # Connection Timed Out
        523,  # Origin Is Unreachable
        524,  # A Timeout Occurred
        525,  # SSL Handshake Failed
        526,  # Invalid SSL Certificate
        527,  # Railgun Error
        530,  # Origin DNS Error
        561,  # Unauthorized
    ]

    def __repr__(self):
        """
        String Representation

        Returns
        -------
        str
        """
        return f"<{self.__class__.__name__}>"

    def __init__(self):
        """
        Initialize with a session
        """
        _user_agent = UserAgent(browsers=["chrome"]).random
        self.session = requests.Session()
        self.headers = {"User-Agent": _user_agent}
        self.session.headers = self.headers
        self.json_headers = self.headers.copy()
        self.json_headers.update({"Content-Type": "application/json"})

    @classmethod
    def get_search_months(cls, search_days) -> List[datetime]:
        """
        Get the Unique Months that need to be Searched

        Returns
        -------
        search_months: List[datetime]
            Datetime Months to search for reservations
        """
        truncated_months = {day.replace(day=1) for day in search_days}
        if len(truncated_months) > 1:
            logger.info(
                f"{len(truncated_months)} different months selected for search, "
                f"ranging from {min(search_days)} to {max(search_days)}"
            )
            return sorted(truncated_months)
        elif len(truncated_months) == 0:
            logger.error(SearchConfig.ERROR_MESSAGE)
            raise RuntimeError(SearchConfig.ERROR_MESSAGE)
        else:
            return sorted(truncated_months)

    @abstractmethod
    def find_campgrounds(self) -> List[CampgroundFacility]:
        """
        List Recreation Areas for the provider
        """

    def make_http_request(
        self,
        url: str,
        method: str = "GET",
        data: Optional[Union[Dict[str, Any], str]] = None,
        headers: Optional[Dict[str, Any]] = None,
        retry_response_codes: Optional[List[int]] = None,
    ) -> requests.Response:
        """
        Make an HTTP Request

        Parameters
        ----------
        url: str
            URL to make the request to
        method: str
            HTTP Method to use. Defaults to GET
        data: Optional[Union[Dict[str, Any], str]]
            Data to send with the request
        headers: Optional[Dict[str, Any]]
            Headers to send with the request
        retry_response_codes: Optional[List[int]]
            List of response codes to raise a ProviderError. on. Defaults to 500 range

        Returns
        -------
        response: requests.Response

        Raises
        ------
        ProviderError
            If the response code is in the retry_response_codes list
        HTTPError
            If the response code is not in the retry_response_codes list and the request fails
        """
        if retry_response_codes is None:
            retry_response_codes = self.FIVE_HUNDRED_STATUS_CODES
        response = self.session.request(
            method=method, url=url, data=data, headers=headers
        )
        if response.status_code not in retry_response_codes:
            response.raise_for_status()
        else:
            error_message = f"HTTP Error - {self.__class__.__name__} - {response.url} - {response.status_code}"
            logger.warning(error_message)
            error_message += f": {response.text}"
            raise ProviderError(error_message)
        return response

    def make_http_request_retry(
        self,
        url: str,
        method: str = "GET",
        data: Optional[Union[Dict[str, Any], str]] = None,
        headers: Optional[Dict[str, Any]] = None,
        retry_response_codes: Optional[List[int]] = None,
    ) -> requests.Response:
        """
        Make an HTTP Request with Exponential Backoff

        This method will retry the request with exponential backoff if the request fails.
        By default, it will only ignore 500 range status codes, but this can be overridden.

        Parameters
        ----------
        url: str
            URL to make the request to
        method: str
            HTTP Method to use. Defaults to GET
        data: Optional[Union[Dict[str, Any], str]]
            Data to send with the request
        headers: Optional[Dict[str, Any]]
            Headers to send with the request
        retry_response_codes: Optional[List[int]]
            List of response codes to retry on. Defaults to 500 range

        Returns
        -------
        response: requests.Response
        """
        retryer = tenacity.Retrying(
            wait=tenacity.wait_random_exponential(
                multiplier=self.RETRY_CONFIG.RETRY_API_MULTIPLIER,
                max=self.RETRY_CONFIG.RETRY_MAX_API_ATTEMPTS,
            ),
            stop=tenacity.stop.stop_after_delay(
                self.RETRY_CONFIG.RETRY_MAX_API_TIMEOUT
            ),
            retry=tenacity.retry_if_exception_type(ProviderError),
        )
        response: requests.Response = retryer.__call__(
            fn=self.make_http_request,
            url=url,
            method=method,
            data=data,
            headers=headers,
            retry_response_codes=retry_response_codes,
        )
        return response



"""
Silent Notifications
"""

import logging
from pprint import pformat
from typing import List

from camply.containers import AvailableCampsite
from camply.notifications.base_notifications import BaseNotifications

logger = logging.getLogger(__name__)


class SilentNotifications(BaseNotifications):
    """
    Silent Notifications
    """

    def send_message(self, message: str, **kwargs) -> None:
        """
        Send a message via Email

        Parameters
        ----------
        message: str
            Email Body
        **kwargs
            kwargs are disregarded

        Returns
        -------
        None
        """
        logger.debug(f"SilentNotification: {message}")

    def send_campsites(self, campsites: List[AvailableCampsite], **kwargs):
        """
        Send a message with a campsite object

        Parameters
        ----------
        campsites: List[AvailableCampsite]
        """
        for campsite in campsites:
            campsite_tuple = (
                (
                    f"{campsite.booking_date.strftime('%Y-%m-%d')} - "
                    f"{campsite.booking_end_date.strftime('%Y-%m-%d')}"
                ),
                campsite.campsite_type,
                campsite.campsite_site_name,
                campsite.recreation_area,
                campsite.facility_name,
                campsite.booking_url,
            )
            message_string = "\n\t• " + "\n\t• ".join(campsite_tuple)
            self.send_message(message_string)
            campsite_formatted = pformat(campsite.dict())
            logger.debug("Campsite Info: " + campsite_formatted)



"""
CLI Testing: `camply list-campsites...`
"""

import pathlib

from pytest import MonkeyPatch

from camply.providers import ReserveCalifornia
from tests.conftest import CamplyRunner, cli_status_checker, vcr_cassette


@vcr_cassette
def test_list_campsites_recdotgov(cli_runner: CamplyRunner) -> None:
    """
    List campsites from Recreation.gov
    """
    test_command = "camply list-campsites --campground 232446"
    result = cli_runner.run_camply_command(command=test_command)
    assert "Wawona Campground" in result.output
    assert "⛺️ 003 - (#540)" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_list_campsites_reservecalifornia(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    List campsites from ReserveCalifornia
    """
    monkeypatch.setattr(ReserveCalifornia, "offline_cache_dir", tmp_path)
    test_command = "camply list-campsites --campground 598 --provider ReserveCalifornia"
    result = cli_runner.run_camply_command(command=test_command)
    assert "Campground Northern End (sites 44-111)" in result.output
    assert "Premium Campsite #82 - (#43467)" in result.output
    cli_status_checker(result=result, exit_code_zero=True)



"""
API Searching Configuration
"""

from os import getenv
from typing import Any, Dict, List, Tuple, Union

from dotenv import load_dotenv

from camply.config.data_columns import DataColumns
from camply.config.file_config import FileConfig
from camply.containers import CampgroundFacility

load_dotenv(FileConfig.DOT_CAMPLY_FILE, override=False)

STANDARD_HEADERS: Dict[str, str] = {
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "en-US,en;q=0.9,la;q=0.8",
    "Cache-Control": "max-age=0",
    "Connection": "keep-alive",
    "Upgrade-Insecure-Requests": "1",
}


class APIConfig:
    """
    Base API Configuration
    """

    RETRY_API_MULTIPLIER: int = 3  # Exponential Backoff Multiplier
    RETRY_MAX_API_ATTEMPTS: int = 10  # Max Number of Retries
    RETRY_MAX_API_TIMEOUT: int = (
        6000  # Max Timeout in Seconds of Retrying (100 Minutes)
    )


class RIDBConfig(APIConfig):
    """
    RIDB API Configuration

    https://ridb.recreation.gov/docs
    """

    _camply_ridb_service_account_api_token: bytes = (
        b"YTc0MTY0NzEtMWI1ZC00YTY0LWFkM2QtYTIzM2U3Y2I1YzQ0"
    )
    _api_key: Union[str, bytes] = getenv(
        "RIDB_API_KEY", _camply_ridb_service_account_api_token
    )
    API_KEY = _camply_ridb_service_account_api_token if _api_key == "" else _api_key

    RIDB_SCHEME: str = "https"
    RIDB_NET_LOC: str = "ridb.recreation.gov"
    RIDB_BASE_PATH: str = "api/v1/"

    # FACILITIES_API_PATH FIELDS
    FACILITIES_API_PATH: str = "facilities"
    CAMPGROUND_FACILITY_FIELD_QUALIFIER: str = "Campground"
    TICKET_FACILITY_FIELD_QUALIFIER: str = "Ticket Facility"
    TIMED_ENTRY_FACILITY_FIELD_QUALIFIER: str = "Timed Entry"
    # RECREATION AREA FIELDS
    REC_AREA_API_PATH: str = "recareas"
    # CAMPSITE DETAILS
    CAMPSITE_API_PATH: str = "campsites"
    # TOUR DETAILS
    TOUR_API_PATH: str = "tours"


class RecreationBookingConfig(APIConfig):
    """
    Variable Storage Class for Recreation.gov Booking API
    """

    API_SCHEME: str = "https"
    API_NET_LOC = "www.recreation.gov"
    API_BASE_PATH: str = "api/camps/availability/campground/"
    API_MONTH_PATH: str = "month"
    API_REFERRERS: Dict[str, Any] = {"Referer": "https://www.recreation.gov/"}

    CAMPSITE_UNAVAILABLE_STRINGS: list = [
        "Reserved",
        "Not Available",
        "Not Reservable",
        "Not Reservable Management",
        "Not Available Cutoff",
        "Lottery",
        "Open",
        "NYR",
        "Closed",
    ]

    CAMPSITE_LOCATION_LOOP_DEFAULT: str = "Default Loop"
    CAMPSITE_LOCATION_SITE_DEFAULT: str = "Default Site"

    CAMPSITE_BOOKING_URL: str = "https://www.recreation.gov/camping/campsites"

    RATE_LIMITING = (1.01, 1.51)


class UseDirectConfig(APIConfig):
    """
    Reserve California API Configuration
    """

    RDR_PREFIX = "rdr"
    SEARCH_PREFIX = "search"
    CITYPARK_ENDPOINT = f"{RDR_PREFIX}/fd/citypark"
    LIST_PLACES_ENDPOINT = f"{RDR_PREFIX}/fd/places"
    LIST_FACILITIES_ENDPOINT = f"{RDR_PREFIX}/fd/facilities"
    SEARCH_ENDPOINT = f"{CITYPARK_ENDPOINT}/namecontains"
    METADATA_PREFIX = f"{RDR_PREFIX}/{SEARCH_PREFIX}/filters"
    PLACE_ENDPOINT = f"{RDR_PREFIX}/{SEARCH_PREFIX}/place"
    AVAILABILITY_ENDPOINT = f"{RDR_PREFIX}/{SEARCH_PREFIX}/grid"
    DATE_FORMAT = "%m-%d-%Y"


class YellowstoneConfig(DataColumns, APIConfig):
    """
    Variable Storage Class
    """

    API_SCHEME: str = "https"
    API_BASE_ENDPOINT: str = "webapi.xanterra.net"
    API_BASE_PATH: str = "v1/api"
    LODGING_PATH: str = f"/{API_BASE_PATH}/availability/hotels"
    YELLOWSTONE_PARK_PATH: str = "yellowstonenationalparklodges"
    YELLOWSTONE_LODGING_PATH: str = f"{LODGING_PATH}/{YELLOWSTONE_PARK_PATH}"
    CAMPSITE_AVAILABILITY: str = f"{API_BASE_PATH}/availability/rooms"
    YELLOWSTONE_CAMPSITE_AVAILABILITY: str = (
        f"{CAMPSITE_AVAILABILITY}/{YELLOWSTONE_PARK_PATH}"
    )
    YELLOWSTONE_PROPERTY_INFO: str = (
        f"{API_BASE_PATH}/property/rooms/{YELLOWSTONE_PARK_PATH}"
    )
    API_REFERRERS: dict = {
        "Host": "webapi.xanterra.net",
        "Origin": "https://secure.yellowstonenationalparklodges.com",
        "Referer": "https://secure.yellowstonenationalparklodges.com/",
    }

    LODGING_CAMPGROUND_QUALIFIER: str = ":RV"

    # JSON FILTERING
    BOOKING_AVAILABILITY: str = "availability"

    # DATAFRAME FILTERING
    FACILITY_ID: str = "facility_id"
    FACILITY_STATUS: str = "status"
    FACILITY_STATUS_QUALIFIER: str = "OPEN"
    FACILITY_ROOMS: str = "rooms"
    FACILITY_AVAILABLE_QUALIFIER: str = "available"
    FACILITY_HOTEL_CODE: str = "hotelCode"
    FACILITY_ROOM_CODE: str = "roomCode"
    FACILITY_PRICE: str = "price"
    FACILITY_TYPE: str = "type"

    CAMPSITE_DATA_COLUMNS: List[str] = ["facility_id", "facility_name", "booking_url"]

    RATE_CODE: str = "INTERNET"
    LODGING_RATES: str = "rates"
    LODGING_TITLE: str = "title"
    LODGING_BASE_PRICES: str = "mins"
    LODGING_OCCUPANCY_BASE: str = "occupancyBase"
    LODGING_OCCUPANCY_MAX: str = "occupancyMax"
    LODGING_ERROR_MESSAGE: str = "message"

    MINIMUM_POLLING_INTERVAL: int = 45

    WEBUI_ALIAS_ENDPOINT: str = "yellowstonenationalparklodges.com"
    WEBUI_BASE_ENDPOINT: str = "secure.yellowstonenationalparklodges.com"
    WEBUI_BOOKING_PATH: str = "booking/lodging-select"

    YELLOWSTONE_RECREATION_AREA_ID: int = 1
    YELLOWSTONE_RECREATION_AREA_NAME: str = "Yellowstone"
    YELLOWSTONE_RECREATION_AREA_FULL_NAME: str = "Yellowstone National Park"
    YELLOWSTONE_RECREATION_AREA_FORMAL_NAME: str = "Yellowstone National Park, USA"
    YELLOWSTONE_LOOP_NAME: str = "N/A"
    CAMPSITE_AVAILABILITY_STATUS: str = "Available"
    YELLOWSTONE_CAMPGROUND_NAME_REPLACE: Tuple[str, str] = (
        "CG Internet Rate",
        "Campground",
    )

    YELLOWSTONE_TIMEZONE: str = "America/Denver"

    # LODGES:  https://webapi.xanterra.net/v1/api/property/hotels/yellowstonenationalparklodges
    YELLOWSTONE_CAMPGROUNDS: Dict[str, str] = {
        "YLYC:RV": "Canyon Campground",
        "YLYB:RV": "Bridge Bay Campground",
        "YLYG:RV": "Grant Campground",
        "YLYM:RV": "Madison Campground",
        "YLYF:RV": "Fishing Bridge RV Park",
    }

    YELLOWSTONE_CAMPGROUND_OBJECTS: List[CampgroundFacility] = []
    for key, value in YELLOWSTONE_CAMPGROUNDS.items():
        YELLOWSTONE_CAMPGROUND_OBJECTS.append(
            CampgroundFacility(
                recreation_area_id=YELLOWSTONE_RECREATION_AREA_ID,
                recreation_area=YELLOWSTONE_RECREATION_AREA_FORMAL_NAME,
                facility_name=value,
                facility_id=str(key),
            )
        )



"""
Camply Configuration Script
"""

import logging
import sys
from collections import OrderedDict
from os.path import isfile
from time import sleep

import rich
from rich.prompt import Confirm, Prompt

from camply.config import FileConfig

logger = logging.getLogger(__name__)


def double_check(message: str) -> bool:
    """
    Double check if a step should be taken within CLI

    Parameters
    ----------
    message: str
        Message to log in interactive shell
    """
    first_confirmation = Confirm.ask(prompt=message)
    if first_confirmation is True:
        second_confirmation = Confirm.ask(prompt="Are you sure?")
        return second_confirmation
    else:
        logging.info("Okay, skipping")
        return False


def check_dot_camply_file() -> bool:
    """
    Check to see if the `.camply` file already exists

    Return the file existence status

    Returns
    -------
    bool
    """
    if isfile(FileConfig.DOT_CAMPLY_FILE) is True:
        logger.info(
            "Skipping configuration. `.camply` file already exists: "
            f"{FileConfig.DOT_CAMPLY_FILE}"
        )
        return True
    else:
        return False


def generate_configuration() -> OrderedDict:
    """
    Generate the Camply Configuration Config

    Returns
    -------
    OrderedDict
        Dict of configuration values
    """
    config_dict = FileConfig.DOT_CAMPLY_FIELDS.copy()
    for field, field_dict in config_dict.items():
        default_value = field_dict["default"]
        field_note = field_dict["notes"]
        if field_note is not None:
            rich.print(
                f"[bold blue]{field}:[/bold blue] "
                f"[bold green]{field_note}[/bold green]"
            )
        message = f"Enter value for [bold blue]{field}[/bold blue]"
        if default_value != "":
            message += f" (default: `[bold purple]{default_value}[/bold purple]`)"
        logged_input = Prompt.ask(prompt=message)
        config_value = logged_input if logged_input != "" else default_value
        config_dict[field] = config_value
    return config_dict


def write_config_to_file(config_dict: OrderedDict) -> None:
    """
    Write the Configuration Object to a file

    Parameters
    ----------
    config_dict : OrderedDict
        Configuration Object
    """
    string_list = [
        "# CAMPLY CONFIGURATION FILE. ",
        "# SEE https://github.com/juftin/camply/blob/main/docs/examples/example.camply",
        "",
    ]
    for config_key, config_value in config_dict.items():
        string_list.append(f'{config_key}="{config_value}"')
    string_list.append("")
    with open(FileConfig.DOT_CAMPLY_FILE, "w") as file_object:
        file_object.write("\n".join(string_list))
        file_object.seek(0)


def generate_dot_camply_file():
    """
    Perform the larger Dot Camply File Generation
    """
    logger.info("Running camply configuration.")
    logger.info(
        "This process generates a configuration file "
        "(https://github.com/juftin/camply/blob/main/docs/examples/example.camply)"
    )
    logger.info("Do not include quotes around values")
    logger.info(
        "To skip a configuration field or keep it as default, just press <Enter>."
    )
    sleep(1.5)
    if isfile(FileConfig.DOT_CAMPLY_FILE):
        logger.warning(
            f".camply file already exists on this machine: {FileConfig.DOT_CAMPLY_FILE}"
        )
        overwrite = double_check(
            "Would you like to overwrite your "
            "[bold yellow].camply[/bold yellow] "
            "configuration file?"
        )
        if overwrite is False:
            sys.exit(0)
    config = generate_configuration()
    if double_check(
        "Are you ready to publish this to a file at "
        f"[bold yellow]{FileConfig.DOT_CAMPLY_FILE}[/bold yellow]"
    ):
        write_config_to_file(config_dict=config)
        logger.info(f"`.camply` file written to machine: {FileConfig.DOT_CAMPLY_FILE}")



"""
Push Notifications via Apprise
"""

import logging
from typing import List

from camply.config import AppriseConfig
from camply.containers import AvailableCampsite
from camply.notifications.base_notifications import BaseNotifications

logger = logging.getLogger(__name__)
logging.getLogger("apprise").setLevel(logging.ERROR)


class AppriseNotifications(BaseNotifications):
    """
    Push Notifications via Apprise
    """

    def __init__(self):
        super().__init__()
        try:
            import apprise
        except ImportError as ie:
            raise RuntimeError(
                "Looks like `apprise` isn't installed. Install it with `pip install camply[apprise]`"
            ) from ie

        if any(
            [
                AppriseConfig.APPRISE_URL is None,
            ]
        ):
            warning_message = (
                "Apprise is not configured properly. To send Apprise notifications "
                "make sure to run `camply configure` or set the "
                "proper environment variable: `APPRISE_URL`."
            )
            logger.error(warning_message)
            raise EnvironmentError(warning_message)
        self.client = apprise.Apprise()
        self.client.add(AppriseConfig.APPRISE_URL)
        logger.info("Apprise: will notify specified URL")

    def send_message(self, message: str, **kwargs):
        """
        Send a message via Apprise - if environment variables are configured

        Parameters
        ----------
        message: str
        """
        self.client.notify(
            body=message,
            title="Camply Notification",
        )

    def send_campsites(self, campsites: List[AvailableCampsite], **kwargs):
        """
        Send a message with a campsite object

        Parameters
        ----------
        campsites: AvailableCampsite
        """
        for campsite in campsites:
            message_title, formatted_dict = self.format_standard_campsites(
                campsite=campsite,
            )
            fields = [f"🏕{message_title}", ""]
            for key, value in formatted_dict.items():
                fields.append(f"{key}: {value}")
            fields.append("")
            fields.append("camply, the campsite finder ⛺️")
            composed_message = "\n".join(fields)
            self.send_message(message=composed_message)



"""
Recreation.gov Implementation for Tours.
"""

import json
import logging
from abc import ABC, abstractmethod
from datetime import date, datetime, time, timedelta, timezone
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
import requests

from camply.config import RecreationBookingConfig, RIDBConfig
from camply.containers import AvailableCampsite
from camply.containers.api_responses import (
    RecDotGovSearchResponse,
    RecDotGovSearchResult,
    TourDailyAvailabilityResponse,
    TourMonthlyAvailabilityResponse,
    TourResponse,
)
from camply.containers.base_container import CamplyModel, RecDotGovEquipment
from camply.providers.base_provider import ProviderSearchError
from camply.providers.recreation_dot_gov.recdotgov_provider import RecreationDotGovBase
from camply.utils import api_utils

logger = logging.getLogger(__name__)


class RecreationDotGovTours(RecreationDotGovBase, ABC):
    """
    Recreation.gov Implementation for Tours
    """

    resource_api_path = RIDBConfig.TOUR_API_PATH
    api_response_class = TourResponse
    api_search_result_class = RecDotGovSearchResult
    api_search_result_key = "entity_id"
    activity_name = None  # Activity Name Should't Be Propogated to Query Parameters

    @property
    @abstractmethod
    def api_search_fq(self) -> str:
        """
        API Query Parameters
        """
        pass

    @property
    @abstractmethod
    def booking_url(self) -> str:
        """
        API Endpoint
        """
        pass

    def paginate_recdotgov_campsites(
        self, facility_id: int, equipment: Optional[List[str]] = None
    ) -> List[RecDotGovSearchResult]:
        """
        Paginate through the RecDotGov Campsite Metadata
        """
        results = 0
        continue_paginate = True
        endpoint_url = api_utils.generate_url(
            scheme=RecreationBookingConfig.API_SCHEME,
            netloc=RecreationBookingConfig.API_NET_LOC,
            path="api/search",
        )
        fq_list = [
            f"asset_id:{facility_id}",
            # Currently, entity_type:tour (parent is entity_type:ticketfacility)
            # or entity_type:timedentry_tour (parent is entity_type:timedentry).
            self.api_search_fq,
        ]
        params = {
            "start": 0,
            "size": 1000,
            "fq": fq_list,
        }
        campsites: List[RecDotGovSearchResult] = []
        while continue_paginate is True:
            response = self.make_recdotgov_request_retry(
                method="GET",
                url=endpoint_url,
                params=params,
            )
            returned_data = json.loads(response.content)
            campsite_response = RecDotGovSearchResponse(**returned_data)
            campsites += campsite_response.results
            results += campsite_response.size
            params.update(start=results)
            if results == campsite_response.total:
                continue_paginate = False
        return campsites

    def make_recdotgov_availability_request(
        self,
        campground_id: int,
        month: datetime,
    ) -> requests.Response:
        """
        Make a request to the RecreationDotGov API

        Parameters
        ----------
        campground_id
        month

        Returns
        -------
        requests.Response
        """
        api_endpoint = self._rec_availability_get_endpoint(
            path=f"{campground_id}/monthlyAvailabilitySummaryView"
        )
        query_params = {
            "year": month.strftime("%Y"),
            "month": month.strftime("%m"),
            "inventoryBucket": "FIT",
        }
        return self.make_recdotgov_request(
            method="GET",
            url=api_endpoint,
            params=query_params,
        )

    @classmethod
    def make_campsite_availability_fields(
        cls,
        tour_id: int,
        booking_url_vars: Dict[str, str],
        booking_date: datetime.date,
        campsite_metadata: pd.DataFrame,
    ) -> Dict[str, Any]:
        """
        Generate a dictionary of fields to be used in a campsite container.

        Parameters
        ----------
        tour_id: int
        booking_url_vars: Dict[str, str]
        booking_date: datetime.date
        campsite_metadata: pd.DataFrame

        Returns
        -------
        Dict[str, Any]
        """
        booking_date = datetime.combine(booking_date, time.min)
        try:
            site_name = campsite_metadata.at[tour_id, "name"]
        except LookupError:
            site_name = f"Tour #{tour_id}"
        try:
            loop_name = campsite_metadata.at[tour_id, "description"]
        except LookupError:
            loop_name = "Description not available"
        try:
            use_type = campsite_metadata.at[tour_id, "time_zone"]
        except LookupError:
            use_type = "Time zone not available"
        return {
            "booking_url": cls.booking_url.format(**booking_url_vars),  # type: ignore
            "booking_date": booking_date,
            "booking_end_date": booking_date + timedelta(days=1),
            "booking_nights": 1,
            "campsite_id": tour_id,
            "campsite_site_name": site_name,
            "campsite_loop_name": loop_name,
            "campsite_type": cls.facility_type,
            "campsite_use_type": use_type,
        }

    @classmethod
    def process_campsite_availability(
        cls,
        availability: Dict[str, Any],
        recreation_area: str,
        recreation_area_id: int,
        facility_name: str,
        facility_id: int,
        month: datetime,
        campsite_metadata: pd.DataFrame,
    ) -> List[Optional[AvailableCampsite]]:
        """
        Parse the JSON Response and return availabilities

        Parameters
        ----------
        availability: dict
            API Response
        recreation_area: str
            Name of Recreation Area
        recreation_area_id: int
            ID of Recreation Area
        facility_name: str
            Campground Facility Name
        facility_id: int
            Campground Facility ID
        month: datetime
            Month to Process
        campsite_metadata: pd.DataFrame
            Metadata Fetched from the Recreation.gov API about the Campsites

        Returns
        -------
        total_campsite_availability: List[Optional[AvailableCampsite]]
            Any monthly availabilities
        """
        total_campsite_availability: List[Optional[AvailableCampsite]] = []
        campsite_data = TourMonthlyAvailabilityResponse(**availability)
        for (
            matching_date,
            date_related_data,
        ) in campsite_data.facility_availability_summary_view_by_local_date.items():
            for (
                tour_id,
                availability_status,
            ) in date_related_data.tour_availability_summary_view_by_tour_id.items():
                if availability_status.reservable > 0:
                    fields = cls.make_campsite_availability_fields(
                        tour_id,
                        vars(availability_status),
                        matching_date,
                        campsite_metadata,
                    )
                    available_campsite = AvailableCampsite(
                        campsite_occupancy=(1, availability_status.reservable),
                        availability_status=availability_status.availability_level,
                        recreation_area=recreation_area,
                        recreation_area_id=recreation_area_id,
                        facility_name=facility_name,
                        facility_id=facility_id,
                        permitted_equipment=[],
                        campsite_attributes=[],
                        **fields,
                    )
                    total_campsite_availability.append(available_campsite)
        return total_campsite_availability

    def get_campground_ids_by_campsites(
        self, campsite_ids: List[int]
    ) -> Tuple[List[int], List[CamplyModel]]:
        """
        Retrieve a list of FacilityIDs, and Facilities from a Campsite ID List

        This implementation for tours tolerates unknown campsite_ids and mock them using known ones.

        Parameters
        ----------
        campsite_ids: List[int]
            List of Campsite IDs

        Returns
        -------
        Tuple[List[int], List[CamplyModel]]
        """
        campground_ids = []
        campgrounds = []
        unknown_ids = []
        for campsite_id in campsite_ids:
            try:
                campsite = self.get_campsite_by_id(campsite_id=campsite_id)
            except ProviderSearchError as e:
                warning_message = (
                    "Ignoring ProviderSearchError; "
                    f"be sure that this is covered by another one in the same facility: {e}"
                )
                logging.warning(warning_message)
                unknown_ids.append(campsite_id)
                continue
            campgrounds.append(campsite)
            campground_ids.append(campsite.FacilityID)
        campground_ids_unique = list(set(campground_ids))
        if unknown_ids:
            if not campground_ids_unique:
                raise ProviderSearchError(
                    "No facility can be determined from specified tours."
                )
            for unknown_id in unknown_ids:
                campsite = TourResponse(
                    TourID=unknown_id,
                    FacilityID=campground_ids_unique[0],
                    TourName=f"Tour #{unknown_id} (facility may be wrong)",
                    TourType="Unknown Tour",
                    TourDuration=0,
                    TourDescription=f"Unknown Tour #{unknown_id}",
                    TourAccessible=False,
                    CreatedDate=date.min,
                    LastUpdatedDate=date.min,
                    ATTRIBUTES=[],
                )
                campgrounds.append(campsite)
        return campground_ids_unique, list(campgrounds)


class RecreationDotGovDailyMixin(RecreationDotGovTours, ABC):
    """
    MixIn Class to Support Daily Searches in Recreation.gov Searches
    """

    @classmethod
    def get_search_months(cls, search_days: List[datetime]) -> List[datetime]:
        """
        Get the Unique Months that need to be Searched

        Returns
        -------
        search_months: List[datetime]
            Datetime Months to search for reservations
        """
        return search_days

    def make_recdotgov_availability_request(
        self,
        campground_id: int,
        month: datetime,
    ) -> requests.Response:
        """
        Make a request to the RecreationDotGov API

        Parameters
        ----------
        campground_id
        month

        Returns
        -------
        requests.Response
        """
        api_endpoint = self._rec_availability_get_endpoint(path=str(campground_id))
        query_params = {
            "date": month.strftime("%Y-%m-%d"),
        }
        return self.make_recdotgov_request(
            method="GET",
            url=api_endpoint,
            params=query_params,
        )

    @classmethod
    def process_campsite_availability(
        cls,
        availability: Dict[str, Any],
        recreation_area: str,
        recreation_area_id: int,
        facility_name: str,
        facility_id: int,
        month: datetime,
        campsite_metadata: pd.DataFrame,
    ) -> List[Optional[AvailableCampsite]]:
        """
        Parse the JSON Response and return availabilities

        Parameters
        ----------
        availability: dict
            API Response
        recreation_area: str
            Name of Recreation Area
        recreation_area_id: int
            ID of Recreation Area
        facility_name: str
            Campground Facility Name
        facility_id: int
            Campground Facility ID
        month: datetime
            Month to Process
        campsite_metadata: pd.DataFrame
            Metadata Fetched from the Recreation.gov API about the Campsites

        Returns
        -------
        total_campsite_availability: List[Optional[AvailableCampsite]]
            Any monthly availabilities
        """
        total_campsite_availability: List[AvailableCampsite] = []
        now = datetime.now(timezone.utc)
        availabilities: Dict[str, Any] = {}
        for slot in availability:
            slot_data = TourDailyAvailabilityResponse(**slot)
            tour_key = (slot_data.tour_date, slot_data.tour_id)
            count_keys = set(slot_data.inventory_count.keys()) & set(
                slot_data.reservation_count.keys()
            )
            for count_key in count_keys:
                window = None
                if "_SECONDARY" in count_key:
                    window = slot_data.booking_windows.SECONDARY
                else:
                    window = slot_data.booking_windows.PRIMARY
                if (
                    not window
                    or now < window.open_timestamp
                    or now >= window.close_timestamp
                ):
                    continue
                inventory_count = slot_data.inventory_count[count_key]
                reservation_count = slot_data.reservation_count[count_key]
                if inventory_count <= reservation_count:
                    continue
                tour_data = availabilities.setdefault(tour_key, {"": slot_data})
                tour_data[slot_data.tour_time] = (
                    tour_data.get(slot_data.tour_time, 0)
                    + inventory_count
                    - reservation_count
                )
        for tour_date, tour_id in availabilities:
            tour_data = availabilities[tour_date, tour_id]
            slot_data = tour_data.pop("")
            fields = cls.make_campsite_availability_fields(
                tour_id,
                vars(slot_data),
                tour_date,
                campsite_metadata,
            )
            available_campsite = AvailableCampsite(
                campsite_occupancy=(1, sum(tour_data.values())),
                availability_status=slot_data.status,
                recreation_area=recreation_area,
                recreation_area_id=recreation_area_id,
                facility_name=facility_name,
                facility_id=facility_id,
                permitted_equipment=[
                    RecDotGovEquipment(
                        equipment_name=tour_time,
                        max_length=available_count,
                    )
                    for tour_time, available_count in tour_data.items()
                ],
                campsite_attributes=[],
                **fields,
            )
            total_campsite_availability.append(available_campsite)
        return total_campsite_availability


class RecreationDotGovTicket(RecreationDotGovTours):
    """
    RecreationDotGovTicket

    Tickets for Tours
    """

    facility_type = RIDBConfig.TICKET_FACILITY_FIELD_QUALIFIER
    api_search_fq = "entity_type:tour"
    api_base_path = "api/ticket/availability/facility/"
    booking_url = "https://www.recreation.gov/ticket/{facility_id}/ticket/{tour_id}"


class RecreationDotGovTimedEntry(RecreationDotGovTours):
    """
    RecreationDotGovTimedEntry

    Timed Entries
    """

    facility_type = RIDBConfig.TIMED_ENTRY_FACILITY_FIELD_QUALIFIER
    api_search_fq = "entity_type:timedentry_tour"
    api_base_path = "api/timedentry/availability/facility/"
    booking_url = (
        "https://www.recreation.gov/timed-entry/{facility_id}/ticket/{tour_id}"
    )


class RecreationDotGovDailyTicket(RecreationDotGovDailyMixin, RecreationDotGovTicket):
    """
    RecreationDotGovTicket: Daily

    Daily MixIn for Tickets
    """


class RecreationDotGovDailyTimedEntry(
    RecreationDotGovDailyMixin, RecreationDotGovTimedEntry
):
    """
    RecreationDotGovTimedEntry: Daily

    Daily MixIn for Tours
    """



"""
ReserveCalifornia Testing
"""

import datetime
import os
import pathlib
from unittest import mock

from dateutil.relativedelta import relativedelta
from pytest import MonkeyPatch

from camply.providers import ReserveCalifornia
from tests.conftest import CamplyRunner, cli_status_checker, vcr_cassette


@vcr_cassette
def test_rc_search_campgrounds_no_api(tmp_path: pathlib.Path) -> None:
    """
    Cached Results: Search Campgrounds
    """
    prov = ReserveCalifornia()
    prov.__offline_cache_dir__ = tmp_path
    expected_facility = 572
    campgrounds = prov.find_campgrounds(rec_area_id=[678])
    found = False
    for campground in campgrounds:
        if campground.facility_name == "North Shore Group Camp":
            assert campground.facility_id == expected_facility
            found = True
    assert found is True


@vcr_cassette
def test_rc_get_campsites(tmp_path: pathlib.Path) -> None:
    """
    Get Campsites
    """
    prov = ReserveCalifornia()
    prov.__offline_cache_dir__ = tmp_path
    start_date = datetime.date(2023, 6, 5)
    campsites = prov.get_campsites(
        campground_id=543,
        start_date=start_date,
        end_date=start_date + relativedelta(days=2),
    )
    assert len({item.campsite_id for item in campsites}) > 1
    assert any(item.campsite_site_name == "Campsite #M93" for item in campsites)


@vcr_cassette
def test_rc_get_metadata(tmp_path: pathlib.Path) -> None:
    """
    Cached Results: Metadata Fetching
    """
    prov = ReserveCalifornia()
    prov.__offline_cache_dir__ = tmp_path
    prov.refresh_metadata()


@vcr_cassette
def test_rc_search_recreation_area(tmp_path: pathlib.Path) -> None:
    """
    Cached Results: Search For RecArea
    """
    prov = ReserveCalifornia()
    prov.__offline_cache_dir__ = tmp_path
    results = prov.search_for_recreation_areas(query="Half Moon Bay")
    assert results[0].recreation_area_location == "Half Moon Bay, CA"


@vcr_cassette
def test_rc_search_campgrounds(tmp_path: pathlib.Path) -> None:
    """
    Cached Results: Search For Campground
    """
    prov = ReserveCalifornia()
    prov.__offline_cache_dir__ = tmp_path
    results = prov.find_campgrounds(search_string="Half Moon Bay", rec_area_id=[])
    assert results[0].recreation_area.__contains__("Half Moon Bay")


@vcr_cassette
def test_rc_cli_recreation_areas(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    CLI Testing - recreation-areas
    """
    monkeypatch.setattr(ReserveCalifornia, "offline_cache_dir", tmp_path)
    test_command = """
    camply recreation-areas --provider ReserveCalifornia --search "Los Angeles"
    """
    result = cli_runner.run_camply_command(test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert "Los Angeles State Historic Park, Calabasas, CA" in result.output


@vcr_cassette
def test_rc_cli_campgrounds(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    CLI Testing - campgrounds
    """
    monkeypatch.setattr(ReserveCalifornia, "offline_cache_dir", tmp_path)
    test_command = """
    camply campgrounds --provider ReserveCalifornia --search "Sonoma Coast"
    """
    result = cli_runner.run_camply_command(test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert "Bodega Dunes" in result.output


@vcr_cassette
def test_rc_cli_campsites(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    CLI Testing - Night Filtering
    """
    monkeypatch.setattr(ReserveCalifornia, "offline_cache_dir", tmp_path)
    test_command = """
    camply campsites \
        --provider ReserveCalifornia \
        --start-date 2023-07-01 \
        --end-date 2023-08-01 \
        --rec-area 718 \
        --weekends
    """
    result = cli_runner.run_camply_command(test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert "total sites found in month of July" in result.output
    assert "Sonoma Coast State Park" in result.output


@vcr_cassette
def test_rc_cli_campsites_nights(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    CLI Testing - Night Filtering
    """
    monkeypatch.setattr(ReserveCalifornia, "offline_cache_dir", tmp_path)
    test_command = """
    camply campsites \
      --start-date 2023-06-01 \
      --end-date 2023-07-01  \
      --campground 1121 \
      --provider ReserveCalifornia
    """
    result = cli_runner.run_camply_command(test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert "Valley Oak Loop (sites 85-90)" in result.output
    assert "total sites found in month of June" in result.output
    assert "https://www.reservecalifornia.com" in result.output


@vcr_cassette
@mock.patch.dict(os.environ, {"CAMPLY_PROVIDER": "ReserveCalifornia"})
def test_rc_cli_campgrounds_env_var(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    CLI Testing - campgrounds via env var
    """
    monkeypatch.setattr(ReserveCalifornia, "offline_cache_dir", tmp_path)
    test_command = """
    camply campgrounds --search "Sonoma Coast"
    """
    result = cli_runner.run_camply_command(test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert "Bodega Dunes" in result.output
    assert 'Using Camply Provider: "ReserveCalifornia"' in result.output



"""
CLI Testing - Campsites: `camply campsites...`
"""

import logging
import pathlib

from pytest import MonkeyPatch

import camply.config
from camply.providers import ReserveCalifornia
from tests.conftest import CamplyRunner, cli_status_checker, vcr_cassette

logger = logging.getLogger(__name__)


@vcr_cassette
def test_no_options_passed(cli_runner: CamplyRunner) -> None:
    """
    Search for Campsites with no options - this should fail
    """
    test_command = """
    camply campsites --debug
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "ERROR" in result.output
    cli_status_checker(result=result, exit_code_zero=False)


@vcr_cassette
def test_search_by_recarea(cli_runner: CamplyRunner) -> None:
    """
    Search for Campsites by RecArea
    """
    test_command = """
    camply \
        campsites \
        --rec-area 2991 \
        --start-date 2023-09-15 \
        --end-date 2023-09-17
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Wawona Campground" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_by_campground(cli_runner: CamplyRunner) -> None:
    """
    Search for Campsites by Campground
    """
    test_command = """
    camply \
        campsites \
        --campground 252037 \
        --start-date 2023-09-15 \
        --end-date 2023-09-17
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Sardine Peak Lookout" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_by_yaml(cli_runner: CamplyRunner) -> None:
    """
    Search for Campsites by YAML
    """
    test_command = """
    camply \
        campsites \
        --yaml-config \
        tests/yaml/example_search.yaml
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "YAML File Parsed: example_search.yaml" in result.output
    assert "Rocky Mountain National Park" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_by_another_yaml(cli_runner: CamplyRunner) -> None:
    """
    Search for Campsites by Another YAML!
    """
    test_command = """
    camply \
        campsites \
        --yaml-config \
        tests/yaml/example_campsite_search.yaml
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "YAML File Parsed: example_campsite_search.yaml" in result.output
    assert "Ledgefork" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_by_misspelled_yaml(cli_runner: CamplyRunner) -> None:
    """
    Search for Campsites by YAML (Misspelled Provider)
    """
    test_command = """
    camply \
        campsites \
        --yaml-config \
        tests/yaml/misspelled_search.yaml
    """
    result = cli_runner.run_camply_command(command=test_command)
    cli_status_checker(result=result, exit_code_zero=False)


@vcr_cassette
def test_search_by_lowercase_yaml(cli_runner: CamplyRunner) -> None:
    """
    Search for Campsites by YAML (Lowercase Provider)
    """
    test_command = """
    camply \
        campsites \
        --yaml-config \
        tests/yaml/lowercase_search.yaml
    """
    result = cli_runner.run_camply_command(command=test_command)
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_nights(cli_runner: CamplyRunner) -> None:
    """
    Search Functionality: Nights
    """
    test_command = """
    camply \
        campsites \
        --campground 232045 \
        --start-date 2023-07-15 \
        --end-date 2023-10-01 \
        --nights 5
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Forbes Creek" in result.output
    assert "5 consecutive" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_yellowstone(cli_runner: CamplyRunner) -> None:
    """
    Search Functionality: Yellowstone Provider
    """
    test_command = """
    camply \
        campsites \
        --provider yellowstone \
        --start-date 2023-10-10 \
        --end-date 2023-10-16
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Madison Campground" in result.output
    assert "Searching for Yellowstone Lodging" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_by_campsite_id(cli_runner: CamplyRunner) -> None:
    """
    Search for Campsites by Campsite ID
    """
    test_command = """
    camply \
        campsites \
        --campsite 40107 \
        --start-date 2023-09-15 \
        --end-date 2023-09-17
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Ledgefork" in result.output
    assert "Searching Specific Campsite" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_by_yellowstone_campground(cli_runner: CamplyRunner) -> None:
    """
    Search for Campsites by Campsite ID
    """
    test_command = """
    camply \
        campsites \
        --provider yellowstone \
        --start-date 2023-09-01 \
        --end-date 2023-09-14 \
        --campground YLYF:RV
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "13 booking nights selected for search" in result.output
    assert "Fishing Bridge RV Park" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_by_equipment(cli_runner: CamplyRunner) -> None:
    """
    Search for Campsites by Equipment
    """
    test_command = """
    camply \
        campsites \
        --rec-area 2018 \
        --start-date 2023-09-09 \
        --end-date 2023-09-17 \
        --nights 3 \
        --equipment RV 25
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Stone Cellar Guard Station" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_multiple_campsites(cli_runner: CamplyRunner) -> None:
    """
    Search for Multiple Campsites
    """
    test_command = """
    camply \
        campsites \
        --campsite 84865 \
        --campsite 84001 \
        --start-date 2023-09-27 \
        --end-date 2023-09-28
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Okanogan-Wenatchee National Forest, WA" in result.output
    assert "Searching Specific Campsite: ⛺️ 116" in result.output
    assert "Searching Specific Campsite: ⛺️ 118" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_error_prone_campground(cli_runner: CamplyRunner) -> None:
    """
    Search for a bad campground: 234779

    This campground didn't have a Parent Rec Area at one point
    """
    test_command = """
    camply \
        campsites \
        --start-date 2023-10-01 \
        --end-date 2023-10-02 \
        --campground 234779
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "USDA Forest Service, OR" in result.output
    assert "Haystack Campground" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_persist_json(cli_runner: CamplyRunner) -> None:
    """
    Persisting Campsites Between Searches: JSON
    """
    test_command = """
    camply \
        --debug \
        campsites \
        --campground 232064 \
        --start-date 2023-09-01 \
        --end-date 2023-10-01 \
        --offline-search \
        --offline-search-path test_file.json
    """
    result_1 = cli_runner.run_camply_command(command=test_command)
    result_2 = cli_runner.run_camply_command(command=test_command)
    camply.config.FileConfig.ROOT_DIRECTORY.joinpath("test_file.json").unlink()
    cli_status_checker(result=result_1, exit_code_zero=True)
    cli_status_checker(result=result_2, exit_code_zero=True)
    assert "Campsite search is configured to save offline" in result_1.output
    assert "test_file.json" in result_2.output
    assert "campsites loaded from file" in result_2.output


@vcr_cassette
def test_search_persist_pickle(cli_runner: CamplyRunner) -> None:
    """
    Persisting Campsites Between Searches: Pickle
    """
    test_command = """
    camply \
        --debug \
        campsites \
        --campground 232064 \
        --start-date 2023-09-01 \
        --end-date 2023-10-01 \
        --offline-search \
        --offline-search-path test_file.pickle
    """
    result_1 = cli_runner.run_camply_command(command=test_command)
    result_2 = cli_runner.run_camply_command(command=test_command)
    camply.config.FileConfig.ROOT_DIRECTORY.joinpath("test_file.pickle").unlink()
    cli_status_checker(result=result_1, exit_code_zero=True)
    cli_status_checker(result=result_2, exit_code_zero=True)
    assert "Campsite search is configured to save offline" in result_1.output
    assert "test_file.pickle" in result_2.output
    assert "campsites loaded from file" in result_2.output


@vcr_cassette
def test_search_once_failure(cli_runner: CamplyRunner) -> None:
    """
    Run Once + Continuous
    """
    test_command = """
    camply campsites \
        --rec-area 2725 \
        --start-date 2023-07-10 \
        --end-date 2023-07-18 \
        --notifications pushover \
        --search-once \
        --continuous
    """
    result = cli_runner.run_camply_command(command=test_command)
    cli_status_checker(result=result, exit_code_zero=False)
    error_message = "You cannot specify `--search-once` alongside `--continuous`"
    assert error_message in result.output


@vcr_cassette
def test_search_once_pushover(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    Run Once + Continuous
    """
    monkeypatch.setattr(ReserveCalifornia, "offline_cache_dir", tmp_path)
    test_command = """
    camply campsites \
        --campground 343 \
        --start-date 2023-07-13 \
        --end-date 2023-07-14 \
        --provider ReserveCalifornia \
        --search-once \
        --notifications pushover
    """
    result = cli_runner.run_camply_command(command=test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert (
        "Notifications active via: <SilentNotifications>, <PushoverNotifications>"
        in result.output
    )
    assert "1 New Campsites Found." in result.output


@vcr_cassette
def test_multiple_search_windows_recdotgov(cli_runner: CamplyRunner) -> None:
    """
    Multiple Search Windows: RecreationDotGov
    """
    test_command = """
    camply campsites \
        --campground 232045 \
        --start-date 2023-07-13 \
        --end-date 2023-07-14 \
        --start-date 2023-08-13 \
        --end-date 2023-08-14
    """
    result = cli_runner.run_camply_command(command=test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert "2 different months selected for search" in result.output
    assert "2 booking nights selected for search" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output


@vcr_cassette
def test_multiple_search_windows_yellowstone(cli_runner: CamplyRunner) -> None:
    """
    Multiple Search Windows: Yellowstone
    """
    test_command = """
    camply campsites \
        --provider yellowstone \
        --start-date 2023-07-13 \
        --end-date 2023-07-14 \
        --start-date 2023-08-13 \
        --end-date 2023-08-14
    """
    result = cli_runner.run_camply_command(command=test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert "2 different months selected for search" in result.output
    assert "2 booking nights selected for search" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output


@vcr_cassette
def test_multiple_search_windows_reservecalifornia(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    Multiple Search Windows: ReserveCalifornia
    """
    monkeypatch.setattr(ReserveCalifornia, "offline_cache_dir", tmp_path)
    test_command = """
    camply campsites \
        --provider ReserveCalifornia \
        --campground 343 \
        --start-date 2023-07-13 \
        --end-date 2023-07-14 \
        --start-date 2023-08-13 \
        --end-date 2023-08-14
    """
    result = cli_runner.run_camply_command(command=test_command)
    cli_status_checker(result=result, exit_code_zero=True)
    assert "2 different months selected for search" in result.output
    assert "2 booking nights selected for search" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output


@vcr_cassette
def test_multiple_search_windows_goingtocamp(cli_runner: CamplyRunner) -> None:
    """
    Multiple Search Windows: GoingToCamp
    """
    test_command = """
    camply \
      campsites \
      --provider goingtocamp \
      --rec-area 1 \
      --start-date 2023-09-01 \
      --end-date 2023-09-02 \
      --start-date 2023-10-01 \
      --end-date 2023-10-02 \
      --equipment-id -32768 \
      --campground -2147483643
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "2 different months selected for search" in result.output
    assert "2 booking nights selected for search" in result.output
    assert "Long Point Region" in result.output
    assert "Waterford North Conservation Area" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_multiple_search_windows_yaml(cli_runner: CamplyRunner) -> None:
    """
    Multiple Search Windows: YAML
    """
    test_command = """
    camply \
        campsites \
        --yaml-config \
        tests/yaml/double_window_campsite_search.yaml
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "2 different months selected for search" in result.output
    assert "2 booking nights selected for search" in result.output
    assert "YAML File Parsed: double_window_campsite_search.yaml" in result.output
    assert "Ledgefork" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_specific_weekdays(cli_runner: CamplyRunner) -> None:
    """
    Only search for specific weekdays
    """
    test_command = """
    camply campsites \
        --campground 232446 \
        --start-date 2023-05-01 \
        --end-date 2023-08-01 \
        --day Wednesday
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "13 booking nights selected for search" in result.output
    assert "3 different months selected for search" in result.output
    assert "Wawona Campground" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_specific_weekdays_yaml(cli_runner: CamplyRunner) -> None:
    """
    Only search for specific weekdays using YAML
    """
    test_command = """
    camply \
        campsites \
        --yaml-config \
        tests/yaml/weekday_search.yaml
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "13 booking nights selected for search" in result.output
    assert "3 different months selected for search" in result.output
    assert "Wawona Campground" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_by_campsites_reservecalifornia(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    Search by Campsite: ReserveCalifornia
    """
    monkeypatch.setattr(ReserveCalifornia, "offline_cache_dir", tmp_path)
    test_command = """
    camply campsites \
      --provider ReserveCalifornia \
      --campground 598 \
      --campsite 43476 \
      --start-date 2023-07-04 \
      --end-date 2023-07-06
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Campground Northern End (sites 44-111)" in result.output
    assert "Campsite #91" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_yaml_search_equipment_array(cli_runner: CamplyRunner) -> None:
    """
    Search by Equipment Array: YAML
    """
    test_command = """
    camply campsites \
      --yaml-config \
      tests/yaml/equipment_scenario_1.yaml
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "YAML File Parsed:" in result.output
    assert "Rocky Mountain National Park" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_yaml_search_equipment_single(cli_runner: CamplyRunner) -> None:
    """
    Search by Equipment Single: YAML
    """
    test_command = """
    camply campsites \
      --yaml-config \
      tests/yaml/equipment_scenario_2.yaml
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "YAML File Parsed:" in result.output
    assert "Rocky Mountain National Park" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_search_by_yaml_reservecalifornia(
    cli_runner: CamplyRunner, tmp_path: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """
    Search by Campsite YAML: ReserveCalifornia
    """
    monkeypatch.setattr(ReserveCalifornia, "offline_cache_dir", tmp_path)
    test_command = """
    camply campsites \
      --yaml-config \
      tests/yaml/reserve_california.yaml
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "YAML File Parsed:" in result.output
    assert "Andrew Molera SP" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result, exit_code_zero=True)



"""
Example Containers
"""

import datetime

from camply.containers.data_containers import AvailableCampsite

example_campsite = AvailableCampsite(
    campsite_id=100,
    booking_date=datetime.datetime(2023, 9, 1),
    booking_end_date=datetime.datetime(2023, 9, 2),
    booking_nights=1,
    campsite_site_name="Test Campsite Name",
    campsite_loop_name="A1",
    campsite_type="Test",
    campsite_occupancy=(1, 5),
    campsite_use_type="Test",
    availability_status="Available",
    recreation_area="Test Recreation Area",
    recreation_area_id=20,
    facility_name="Test Campground",
    facility_id=50,
    booking_url="https://youtu.be/eBGIQ7ZuuiU",
    permitted_equipment=[],
    campsite_attributes=[],
)



"""
providers __init__ file
"""

from typing import Union

from .base_provider import BaseProvider
from .going_to_camp.going_to_camp_provider import GoingToCamp
from .recreation_dot_gov.recdotgov_camps import RecreationDotGov
from .recreation_dot_gov.recdotgov_tours import (
    RecreationDotGovDailyTicket,
    RecreationDotGovDailyTimedEntry,
    RecreationDotGovTicket,
    RecreationDotGovTimedEntry,
)
from .usedirect.variations import (
    AlabamaStateParks,
    ArizonaStateParks,
    FairfaxCountyParks,
    FloridaStateParks,
    MaricopaCountyParks,
    MinnesotaStateParks,
    MissouriStateParks,
    NorthernTerritory,
    OhioStateParks,
    OregonMetro,
    ReserveCalifornia,
    VirginiaStateParks,
)
from .xanterra.yellowstone_lodging import Yellowstone

ProviderType = Union[
    GoingToCamp,
    RecreationDotGov,
    RecreationDotGovDailyTicket,
    RecreationDotGovDailyTimedEntry,
    RecreationDotGovTicket,
    RecreationDotGovTimedEntry,
    Yellowstone,
    ReserveCalifornia,
    NorthernTerritory,
    FloridaStateParks,
    OregonMetro,
    OhioStateParks,
    VirginiaStateParks,
    ArizonaStateParks,
    MaricopaCountyParks,
    MissouriStateParks,
    AlabamaStateParks,
    FairfaxCountyParks,
    MinnesotaStateParks,
]

__all__ = [
    "BaseProvider",
    "ProviderType",
    "GoingToCamp",
    "RecreationDotGov",
    "RecreationDotGovDailyTicket",
    "RecreationDotGovDailyTimedEntry",
    "RecreationDotGovTicket",
    "RecreationDotGovTimedEntry",
    "Yellowstone",
    "ReserveCalifornia",
    "NorthernTerritory",
    "FloridaStateParks",
    "OregonMetro",
    "OhioStateParks",
    "VirginiaStateParks",
    "ArizonaStateParks",
    "MaricopaCountyParks",
    "MissouriStateParks",
    "AlabamaStateParks",
    "FairfaxCountyParks",
    "MinnesotaStateParks",
]



"""
CLI Testing: `camply ...`
"""

import logging

from camply import __version__
from camply.cli import camply_command_line
from tests.conftest import CamplyRunner, cli_status_checker, vcr_cassette

logger = logging.getLogger(__name__)


@vcr_cassette
def test_main_succeeds(cli_runner: CamplyRunner) -> None:
    """
    It exits with a status code of zero.
    """
    result = cli_runner.invoke(camply_command_line)
    assert "Welcome to camply, the campsite finder" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_debug(cli_runner: CamplyRunner) -> None:
    """
    Use the Debug Option
    """
    test_command = """
    camply \
        campgrounds \
        --provider Yellowstone \
        --debug
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert __version__ in result.output
    cli_status_checker(result=result)



"""
UseDirect API Responses
"""

import datetime
from typing import Any, Dict, List, Optional

from camply.containers import CamplyModel


class UseDirectRestrictions(CamplyModel):
    """
    UseDirect: Campsite Restrictions
    """

    FutureBookingStarts: Optional[datetime.datetime]
    FutureBookingEnds: Optional[datetime.datetime]
    MinimumStay: Optional[int]
    MaximumStay: Optional[int]
    IsRestrictionValid: Optional[bool]
    Time: Optional[str]


class UseDirectUnitType(CamplyModel):
    """
    UseDirect: Unit Types
    """

    UnitTypeId: int
    UseType: int
    Name: str
    Available: bool
    AvailableFiltered: Optional[bool]
    UnitCategoryId: Optional[int]
    UnitTypeGroupId: Optional[int]
    MaxVehicleLength: Optional[int]
    HasAda: Optional[bool]
    Restrictions: Optional[Any]
    AvailableCount: Optional[int]


class UseDirectAvailabilitySlice(CamplyModel):
    """
    Slice of Availability per Date
    """

    Date: datetime.date
    IsFree: bool
    IsBlocked: Optional[bool]
    IsWalkin: Optional[bool]
    ReservationId: Optional[int]
    Lock: Any
    MinStay: Optional[int]
    IsReservationDraw: Optional[bool]


class UseDirectAvailabilityUnit(CamplyModel):
    """
    Unit of Availability in Availability Request
    """

    UnitId: Optional[int]
    Name: Optional[str]
    ShortName: Optional[str]
    RecentPopups: Optional[int]
    IsAda: Optional[bool]
    AllowWebBooking: Optional[bool]
    MapInfo: Optional[Dict[str, Any]] = {}
    IsWebViewable: Optional[bool]
    IsFiltered: Optional[bool]
    UnitCategoryId: Optional[int]
    SleepingUnitIds: Optional[List[int]]
    UnitTypeGroupId: Optional[int]
    UnitTypeId: Optional[int]
    UseType: Optional[int]
    VehicleLength: Optional[int]
    OrderBy: Optional[int]
    OrderByRaw: Optional[int]
    SliceCount: Optional[int]
    AvailableCount: Optional[int]
    Slices: Optional[Dict[datetime.datetime, UseDirectAvailabilitySlice]] = {}
    StartTime: Any
    EndTime: Any
    FacilityId: Optional[int]


class UseDirectFacility(CamplyModel):
    """
    Campground Representation for UseDirect
    """

    FacilityId: int
    Name: Optional[str]
    Description: Optional[str]
    RateMessage: Optional[str]
    FacilityType: Optional[int]
    FacilityTypeNew: Optional[int]
    InSeason: Optional[bool]
    Available: Optional[bool]
    AvailableFiltered: Optional[bool]
    Restrictions: Optional[UseDirectRestrictions]
    Latitude: Optional[float]
    Longitude: Optional[float]
    Category: Optional[str]
    EnableCheckOccupancy: Optional[bool]
    AvailableOccupancy: Any
    FacilityAllowWebBooking: Optional[bool]
    UnitTypes: Optional[Dict[int, UseDirectUnitType]] = {}
    IsAvailableForGroup: Optional[bool]
    IsAvailableForPatron: Optional[bool]
    IsAvailableForEducationalGroup: Optional[bool]
    IsAvailableForCto: Optional[bool]
    FacilityBehaviourType: Optional[int]


class UseDirectPlace(CamplyModel):
    """
    UseDirect: Place Object
    """

    PlaceId: int
    Name: str
    Description: Optional[str]
    HasAlerts: Optional[bool]
    IsFavourite: Optional[bool]
    Allhighlights: Optional[str]
    Url: Optional[str]
    ImageUrl: Optional[str]
    BannerUrl: Optional[str]
    ParkSize: Optional[str]
    Latitude: Optional[float]
    Longitude: Optional[float]
    TimeZone: Optional[str]
    TimeStamp: Optional[datetime.datetime]
    MilesFromSelected: Optional[int]
    Available: bool
    AvailableFiltered: Optional[bool]
    ParkCategoryId: Optional[int]
    ParkActivity: Optional[int]
    ParkPopularity: Optional[int]
    AvailableUnitCount: Optional[int]
    Restrictions: Optional[UseDirectRestrictions]
    Facilities: Optional[Dict[int, UseDirectFacility]] = {}
    IsAvailableForGreatwalk: Optional[bool]
    FacilityDefaultZoom: Optional[int]


class UseDirectDetailedPlace(CamplyModel):
    """
    https://calirdr.usedirect.com/RDR/rdr/fd/places
    """

    PlaceId: int
    Name: str
    Description: Optional[str]
    ParkSize: Optional[str]
    Latitude: Optional[float]
    Longitude: Optional[float]
    ParkCategoryId: Optional[int]
    ParkActivity: Optional[int]
    ParkPopularity: Optional[int]
    IsAvailableForGreatwalk: Optional[bool]
    FacilityDefaultZoom: Optional[int]
    RegionId: Optional[int]
    ShortName: Optional[str]
    OrderBy: Optional[int]
    AllowWebBooking: Optional[bool]
    InventoryLocking: Optional[bool]
    InventoryLockDuration: Optional[int]
    UsePrepend: Optional[bool]
    PrependCode: Optional[str]
    Address1: Optional[str]
    Address2: Optional[str]
    City: Optional[str]
    State: Optional[str]
    Zip: Optional[str]
    VoicePhone: Optional[str]
    UDate: datetime.datetime
    UserId: Optional[int]
    RowGuid: Optional[str]
    IsWebViewable: Optional[bool]
    WeekdayCheckdays: Optional[int]
    WeekendCheckdays: Optional[int]


class UseDirectAvailabilityFacility(UseDirectFacility):
    """
    UseDirect: Facility w/ Availability
    """

    FacilityMapSize: Optional[bool]
    FacilityImage: Optional[str]
    FacilityImageVBT: Optional[str]
    DatesInSeason: Optional[int]
    DatesOutOfSeason: Optional[int]
    SeasonDates: Optional[Dict[datetime.datetime, bool]] = {}
    TrafficStatuses: Optional[Dict[str, Any]] = {}
    UnitCount: Optional[int]
    AvailableUnitCount: Optional[int]
    SliceCount: Optional[int]
    AvailableSliceCount: Optional[int]
    TimebaseMaxHours: Optional[int]
    TimebaseMinHours: Optional[int]
    TimebaseDuration: Optional[float]
    IsReservationDraw: Optional[bool]
    DrawBookingStartDate: Optional[datetime.datetime]
    DrawBookingEndDate: Optional[datetime.datetime]
    Units: Optional[Dict[str, UseDirectAvailabilityUnit]]


class UseDirectAvailabilityResponse(CamplyModel):
    """
    API Response from /rdr/rdr/search/grid
    """

    Message: str
    Filters: Optional[Dict[str, Any]] = {}
    UnitTypeId: int
    StartDate: Optional[datetime.date]
    EndDate: Optional[datetime.date]
    NightsRequested: Optional[int]
    NightsActual: Optional[int]
    TodayDate: Optional[datetime.date]
    TimeZone: Optional[str]
    TimeStamp: Optional[datetime.datetime]
    MinDate: Optional[datetime.date]
    MaxDate: Optional[datetime.date]
    AvailableUnitsOnly: Optional[bool]
    UnitSort: Optional[str]
    TimeGrid: Optional[bool]
    ForUnit: Optional[bool]
    UnitId: Optional[int]
    TimeBetween: Optional[str]
    TimeBetweenEval: Optional[str]
    Facility: Optional[UseDirectAvailabilityFacility]


class UseDirectUnitCategory(CamplyModel):
    """
    UseDirect: Unit Categories
    """

    UnitCategoryId: int
    UnitCategoryName: str
    HasEquipment: Optional[bool]
    Icon: Optional[str]


class UseDirectNightlySleepingUnit(CamplyModel):
    """
    UseDirect: Nightly Sleeping Units
    """

    UnitCategoryId: int
    SleepingUnitId: int
    SleepingUnitName: str
    IsWheeled: Optional[bool]
    Icon: Optional[str]


class UseDirectMinVehicleLength(CamplyModel):
    """
    UseDirect: Vehicle Length
    """

    SleepingUnitId: int
    MinVehicleLength: int
    MinVehicleName: str
    Icon: Optional[str]


class UseDirectUnitTypeGroup(CamplyModel):
    """
    UseDirect: Unit Type Groups
    """

    UnitCategoryId: int
    UnitTypesGroupId: int
    UnitTypesGroupName: str
    Icon: Optional[str]


class UseDirectAmenity(CamplyModel):
    """
    UseDirect: Amenities
    """

    AmenityId: int
    Name: str
    ShortName: str
    AmenityType: Optional[str]
    IsSearchable: Optional[bool]
    Description: Optional[str]
    OrderBy: Optional[int]
    IDate: Optional[datetime.datetime]
    UDate: Optional[datetime.datetime]
    ImagePath: Optional[str]
    UCashierName: Optional[str]
    UStoreId: Optional[int]
    IsADA: Optional[bool]
    Value: Optional[Any]


class UseDirectMetadata(CamplyModel):
    """
    Campground Metadata Responses
    """

    Message: str
    UnitCategories: Optional[List[UseDirectUnitCategory]] = []
    NightlySleepingUnits: Optional[List[UseDirectNightlySleepingUnit]] = []
    MinVehicleLengths: Optional[List[UseDirectMinVehicleLength]] = []
    UnitTypesGroups: Optional[List[UseDirectUnitTypeGroup]] = []
    PlaceHighlights: Optional[List[Any]] = []
    AllAmenity: Optional[List[UseDirectAmenity]] = []


class UseDirectCityPark(CamplyModel):
    """
    UseDirect: City Parks
    """

    CityParkId: int
    Name: str
    Latitude: float
    Longitude: float
    IsActive: bool
    EntityType: Optional[str]
    EnterpriseId: Optional[int]
    ParkSize: Optional[str]
    PlaceId: int


class UseDirectFacilityMetadata(CamplyModel):
    """
    UseDirect: Facility Metadata
    """

    FacilityId: int
    RegionId: Optional[int]
    PlaceId: int
    Name: str
    ShortName: Optional[str]
    Description: Optional[str]
    OrderBy: Optional[int]
    FacilityType: Optional[int]
    UsePrepend: Optional[bool]
    PrependCode: Optional[str]
    AllowWebBooking: Optional[bool]
    AutoOpenInventory: Optional[bool]
    UDate: Optional[datetime.datetime]
    RowGuid: Optional[str]
    FacilityTypeNew: Optional[int]
    MaxPersonOccupancy: Optional[int]
    IsAvailableForGroup: Optional[bool]
    IsAvailableForEducationalGroup: Optional[bool]
    IsAvailableForCto: Optional[bool]
    IsCaptcha: Optional[bool]
    FacilityBehaviourType: Optional[int]
    EnableCheckOccupancy: Optional[bool]
    IsTrail: Optional[bool]
    TimebaseMaxHours: Optional[int]
    TimebaseMinHours: Optional[int]
    TimebaseDuration: Optional[int]



"""
Push Notifications via ntfy.sh
"""

import logging
from typing import List

import requests

from camply.config import NtfyConfig
from camply.containers import AvailableCampsite
from camply.notifications.base_notifications import BaseNotifications

logger = logging.getLogger(__name__)


class NtfyNotifications(BaseNotifications):
    """
    Push Notifications via Ntfy
    """

    def __init__(self):
        super().__init__()
        if any([NtfyConfig.NTFY_TOPIC is None, NtfyConfig.NTFY_TOPIC == ""]):
            warning_message = (
                "Ntfy is not configured properly. To send Ntfy messages "
                "set the proper environment variable: `NTFY_TOPIC`."
            )
            logger.error(warning_message)
            raise EnvironmentError(warning_message)

    def send_message(self, message: str, **kwargs) -> requests.Response:
        """
        Send a message via Ntfy - if environment variables are configured

        Parameters
        ----------
        message: str

        Returns
        -------
        requests.Response
        """
        response = self.session.post(
            url=NtfyConfig.NTFY_API_ENDPOINT + NtfyConfig.NTFY_TOPIC,
            data=message.encode("utf-8"),
            headers={
                "Title": kwargs.get("title", "Camply Notification"),
                "Click": kwargs.get("url", ""),
            },
        )
        try:
            response.raise_for_status()
        except requests.HTTPError as he:
            logger.warning(
                "Notifications weren't able to be sent to Ntfy. "
                "Your configuration might be incorrect."
            )
            raise ConnectionError(response.text) from he
        return response

    def send_campsites(self, campsites: List[AvailableCampsite], **kwargs):
        """
        Send a message with a campsite object

        Parameters
        ----------
        campsites: AvailableCampsite
        """
        for campsite in campsites:
            message_title, formatted_dict = self.format_standard_campsites(
                campsite=campsite,
            )
            fields = []
            for key, value in formatted_dict.items():
                fields.append(f"{key}: {value}")
            composed_message = "\n".join(fields)
            self.send_message(
                message=composed_message,
                title=message_title,
                url=formatted_dict.get("Booking Link", ""),
            )
